"""Tests for Hamilton Nimbus backend implementation.

This module tests the Nimbus backend, its command classes, tip type handling,
and liquid handling operations.
"""

import unittest
import unittest.mock
from typing import Optional

from pylabrobot.liquid_handling.backends.hamilton.nimbus_backend import (
  Aspirate,
  DisableADC,
  Dispense,
  DropTips,
  DropTipsRoll,
  EnableADC,
  GetChannelConfiguration,
  GetChannelConfiguration_1,
  InitializeSmartRoll,
  IsDoorLocked,
  IsInitialized,
  IsTipPresent,
  LockDoor,
  NimbusBackend,
  NimbusTipType,
  Park,
  PickupTips,
  PreInitializeSmart,
  SetChannelConfiguration,
  UnlockDoor,
  _get_tip_type_from_tip,
)
from pylabrobot.liquid_handling.backends.hamilton.tcp.messages import HoiParams, HoiParamsParser
from pylabrobot.liquid_handling.backends.hamilton.tcp.packets import Address
from pylabrobot.liquid_handling.backends.hamilton.tcp.protocol import HamiltonProtocol
from pylabrobot.liquid_handling.standard import (
  Drop,
  Pickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)
from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.corning.plates import Cor_96_wellplate_360ul_Fb
from pylabrobot.resources.hamilton import HamiltonTip, TipPickupMethod, TipSize
from pylabrobot.resources.hamilton.nimbus_decks import NimbusDeck
from pylabrobot.resources.hamilton.tip_racks import hamilton_96_tiprack_300uL


class TestNimbusTipType(unittest.TestCase):
  """Tests for NimbusTipType enum and tip type mapping."""

  def test_tip_type_values(self):
    self.assertEqual(NimbusTipType.STANDARD_300UL, 0)
    self.assertEqual(NimbusTipType.STANDARD_300UL_FILTER, 1)
    self.assertEqual(NimbusTipType.LOW_VOLUME_10UL, 2)
    self.assertEqual(NimbusTipType.LOW_VOLUME_10UL_FILTER, 3)
    self.assertEqual(NimbusTipType.HIGH_VOLUME_1000UL, 4)
    self.assertEqual(NimbusTipType.HIGH_VOLUME_1000UL_FILTER, 5)
    self.assertEqual(NimbusTipType.TIP_50UL, 22)
    self.assertEqual(NimbusTipType.TIP_50UL_FILTER, 23)
    self.assertEqual(NimbusTipType.SLIM_CORE_300UL, 36)

  def test_get_tip_type_low_volume(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=False,
      total_tip_length=50.0,
      maximal_volume=10.0,
      tip_size=TipSize.LOW_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.LOW_VOLUME_10UL)

  def test_get_tip_type_low_volume_filter(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=True,
      total_tip_length=50.0,
      maximal_volume=10.0,
      tip_size=TipSize.LOW_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.LOW_VOLUME_10UL_FILTER)

  def test_get_tip_type_standard_50ul(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=False,
      total_tip_length=50.0,
      maximal_volume=50.0,
      tip_size=TipSize.STANDARD_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.TIP_50UL)

  def test_get_tip_type_standard_50ul_filter(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=True,
      total_tip_length=50.0,
      maximal_volume=50.0,
      tip_size=TipSize.STANDARD_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.TIP_50UL_FILTER)

  def test_get_tip_type_standard_300ul(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=False,
      total_tip_length=59.9,
      maximal_volume=300.0,
      tip_size=TipSize.STANDARD_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.STANDARD_300UL)

  def test_get_tip_type_standard_300ul_filter(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=True,
      total_tip_length=59.9,
      maximal_volume=300.0,
      tip_size=TipSize.STANDARD_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.STANDARD_300UL_FILTER)

  def test_get_tip_type_high_volume(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=False,
      total_tip_length=95.1,
      maximal_volume=1000.0,
      tip_size=TipSize.HIGH_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.HIGH_VOLUME_1000UL)

  def test_get_tip_type_high_volume_filter(self):
    tip = HamiltonTip(
      name="test_tip",
      has_filter=True,
      total_tip_length=95.1,
      maximal_volume=1000.0,
      tip_size=TipSize.HIGH_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )
    self.assertEqual(_get_tip_type_from_tip(tip), NimbusTipType.HIGH_VOLUME_1000UL_FILTER)

  def test_get_tip_type_non_hamilton_tip_raises(self):
    from pylabrobot.resources import Tip

    # Regular Tip (non-Hamilton) should raise
    tip = Tip(
      name="test_tip",
      has_filter=False,
      total_tip_length=50.0,
      maximal_volume=300.0,
      fitting_depth=8.0,
    )
    with self.assertRaises(ValueError) as ctx:
      _get_tip_type_from_tip(tip)
    self.assertIn("HamiltonTip", str(ctx.exception))


class TestNimbusCommands(unittest.TestCase):
  """Tests for Nimbus command classes."""

  def test_lock_door_command(self):
    cmd = LockDoor(Address(1, 1, 268))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 1)
    self.assertEqual(cmd.protocol, HamiltonProtocol.OBJECT_DISCOVERY)
    # Default action_code is 3 (COMMAND_REQUEST)
    self.assertEqual(cmd.action_code, 3)

  def test_unlock_door_command(self):
    cmd = UnlockDoor(Address(1, 1, 268))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 2)

  def test_is_door_locked_command(self):
    cmd = IsDoorLocked(Address(1, 1, 268))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 3)
    # STATUS_REQUEST must have action_code=0
    self.assertEqual(cmd.action_code, 0)

  def test_is_door_locked_parse_response(self):
    # Simulate response: bool fragment with True
    response_data = HoiParams().bool_value(True).build()
    result = IsDoorLocked.parse_response_parameters(response_data)
    self.assertEqual(result, {"locked": True})

    response_data = HoiParams().bool_value(False).build()
    result = IsDoorLocked.parse_response_parameters(response_data)
    self.assertEqual(result, {"locked": False})

  def test_park_command(self):
    cmd = Park(Address(1, 1, 48896))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 3)

  def test_is_initialized_command(self):
    cmd = IsInitialized(Address(1, 1, 48896))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 14)
    self.assertEqual(cmd.action_code, 0)

  def test_is_initialized_parse_response(self):
    response_data = HoiParams().bool_value(True).build()
    result = IsInitialized.parse_response_parameters(response_data)
    self.assertEqual(result, {"initialized": True})

  def test_is_tip_present_command(self):
    cmd = IsTipPresent(Address(1, 1, 257))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 16)
    self.assertEqual(cmd.action_code, 0)

  def test_is_tip_present_parse_response(self):
    # Simulate response with i16 array
    response_data = HoiParams().i16_array([1, 0, 1, 0, 0, 0, 0, 0]).build()
    result = IsTipPresent.parse_response_parameters(response_data)
    self.assertEqual(result["tip_present"], [1, 0, 1, 0, 0, 0, 0, 0])

  def test_get_channel_configuration_1_command(self):
    cmd = GetChannelConfiguration_1(Address(1, 1, 48896))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 15)
    self.assertEqual(cmd.action_code, 0)

  def test_get_channel_configuration_1_parse_response(self):
    response_data = HoiParams().u16(8).i16_array([0, 0, 0, 0, 0, 0, 0, 0]).build()
    result = GetChannelConfiguration_1.parse_response_parameters(response_data)
    self.assertEqual(result["channels"], 8)
    self.assertEqual(result["channel_types"], [0, 0, 0, 0, 0, 0, 0, 0])

  def test_pre_initialize_smart_command(self):
    cmd = PreInitializeSmart(Address(1, 1, 257))
    self.assertEqual(cmd.interface_id, 1)
    self.assertEqual(cmd.command_id, 32)


class TestNimbusCommandParameters(unittest.TestCase):
  """Tests for Nimbus command parameter building."""

  def test_initialize_smart_roll_parameters(self):
    cmd = InitializeSmartRoll(
      dest=Address(1, 1, 48896),
      x_positions=[1000, 2000, 3000, 4000, 0, 0, 0, 0],
      y_positions=[5000, 6000, 7000, 8000, 0, 0, 0, 0],
      begin_tip_deposit_process=[9000, 9000, 9000, 9000, 0, 0, 0, 0],
      end_tip_deposit_process=[8500, 8500, 8500, 8500, 0, 0, 0, 0],
      z_position_at_end_of_a_command=[14600, 14600, 14600, 14600, 0, 0, 0, 0],
      roll_distances=[900, 900, 900, 900, 0, 0, 0, 0],
    )

    params = cmd.build_parameters()
    data = params.build()

    # Verify parameters were built - should have 6 i32_array fragments
    parser = HoiParamsParser(data)
    results = parser.parse_all()
    self.assertEqual(len(results), 6)

    # Verify x_positions
    self.assertEqual(results[0][1], [1000, 2000, 3000, 4000, 0, 0, 0, 0])
    # Verify y_positions
    self.assertEqual(results[1][1], [5000, 6000, 7000, 8000, 0, 0, 0, 0])

  def test_pickup_tips_parameters(self):
    cmd = PickupTips(
      dest=Address(1, 1, 257),
      channels_involved=[1, 1, 0, 0, 0, 0, 0, 0],
      x_positions=[10000, 11000, 0, 0, 0, 0, 0, 0],
      y_positions=[20000, 21000, 0, 0, 0, 0, 0, 0],
      minimum_traverse_height_at_beginning_of_a_command=14600,
      begin_tip_pick_up_process=[5000, 5000, 0, 0, 0, 0, 0, 0],
      end_tip_pick_up_process=[4500, 4500, 0, 0, 0, 0, 0, 0],
      tip_types=[0, 0, 0, 0, 0, 0, 0, 0],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Should have: u16_array, i32_array, i32_array, i32, i32_array, i32_array, u16_array
    self.assertEqual(len(results), 7)

    # Verify channels_involved (u16 array)
    self.assertEqual(results[0][1], [1, 1, 0, 0, 0, 0, 0, 0])
    # Verify traverse height (i32)
    self.assertEqual(results[3][1], 14600)

  def test_drop_tips_parameters(self):
    cmd = DropTips(
      dest=Address(1, 1, 257),
      channels_involved=[1, 0, 0, 0, 0, 0, 0, 0],
      x_positions=[10000, 0, 0, 0, 0, 0, 0, 0],
      y_positions=[20000, 0, 0, 0, 0, 0, 0, 0],
      minimum_traverse_height_at_beginning_of_a_command=14600,
      begin_tip_deposit_process=[5000, 0, 0, 0, 0, 0, 0, 0],
      end_tip_deposit_process=[4500, 0, 0, 0, 0, 0, 0, 0],
      z_position_at_end_of_a_command=[14600, 0, 0, 0, 0, 0, 0, 0],
      default_waste=False,
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Should have: u16_array, i32_array, i32_array, i32, i32_array, i32_array, i32_array, bool
    self.assertEqual(len(results), 8)
    # Verify default_waste (bool)
    self.assertEqual(results[7][1], False)

  def test_drop_tips_roll_parameters(self):
    cmd = DropTipsRoll(
      dest=Address(1, 1, 257),
      channels_involved=[1, 1, 1, 1, 0, 0, 0, 0],
      x_positions=[10000, 11000, 12000, 13000, 0, 0, 0, 0],
      y_positions=[20000, 21000, 22000, 23000, 0, 0, 0, 0],
      minimum_traverse_height_at_beginning_of_a_command=14600,
      begin_tip_deposit_process=[5000, 5000, 5000, 5000, 0, 0, 0, 0],
      end_tip_deposit_process=[4500, 4500, 4500, 4500, 0, 0, 0, 0],
      z_position_at_end_of_a_command=[14600, 14600, 14600, 14600, 0, 0, 0, 0],
      roll_distances=[900, 900, 900, 900, 0, 0, 0, 0],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Should have: u16_array, i32_array, i32_array, i32, i32_array, i32_array, i32_array, i32_array
    self.assertEqual(len(results), 8)
    # Verify roll_distances
    self.assertEqual(results[7][1], [900, 900, 900, 900, 0, 0, 0, 0])

  def test_enable_adc_parameters(self):
    cmd = EnableADC(
      dest=Address(1, 1, 257),
      channels_involved=[1, 1, 0, 0, 0, 0, 0, 0],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    self.assertEqual(len(results), 1)
    self.assertEqual(results[0][1], [1, 1, 0, 0, 0, 0, 0, 0])

  def test_disable_adc_parameters(self):
    cmd = DisableADC(
      dest=Address(1, 1, 257),
      channels_involved=[1, 1, 1, 1, 0, 0, 0, 0],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    self.assertEqual(len(results), 1)
    self.assertEqual(results[0][1], [1, 1, 1, 1, 0, 0, 0, 0])

  def test_set_channel_configuration_parameters(self):
    cmd = SetChannelConfiguration(
      dest=Address(1, 1, 257),
      channel=1,
      indexes=[1, 3, 4],
      enables=[True, False, False, False],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Should have: u16, i16_array, bool_array
    self.assertEqual(len(results), 3)
    self.assertEqual(results[0][1], 1)  # channel
    self.assertEqual(results[1][1], [1, 3, 4])  # indexes
    self.assertEqual(results[2][1], [True, False, False, False])  # enables

  def test_get_channel_configuration_parameters(self):
    cmd = GetChannelConfiguration(
      dest=Address(1, 1, 257),
      channel=2,
      indexes=[2],
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    self.assertEqual(len(results), 2)
    self.assertEqual(results[0][1], 2)  # channel
    self.assertEqual(results[1][1], [2])  # indexes


class TestAspirateDispenseCommands(unittest.TestCase):
  """Tests for Aspirate and Dispense command parameter building."""

  def test_aspirate_parameters(self):
    cmd = Aspirate(
      dest=Address(1, 1, 257),
      aspirate_type=[0, 0, 0, 0, 0, 0, 0, 0],
      channels_involved=[1, 0, 0, 0, 0, 0, 0, 0],
      x_positions=[10000, 0, 0, 0, 0, 0, 0, 0],
      y_positions=[20000, 0, 0, 0, 0, 0, 0, 0],
      minimum_traverse_height_at_beginning_of_a_command=14600,
      lld_search_height=[1000, 0, 0, 0, 0, 0, 0, 0],
      liquid_height=[500, 0, 0, 0, 0, 0, 0, 0],
      immersion_depth=[300, 0, 0, 0, 0, 0, 0, 0],
      surface_following_distance=[0, 0, 0, 0, 0, 0, 0, 0],
      minimum_height=[0, 0, 0, 0, 0, 0, 0, 0],
      clot_detection_height=[0, 0, 0, 0, 0, 0, 0, 0],
      min_z_endpos=0,
      swap_speed=[1000, 0, 0, 0, 0, 0, 0, 0],
      blow_out_air_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      pre_wetting_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      aspirate_volume=[1000, 0, 0, 0, 0, 0, 0, 0],  # 100uL in 0.1uL units
      transport_air_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      aspiration_speed=[500, 0, 0, 0, 0, 0, 0, 0],
      settling_time=[10, 0, 0, 0, 0, 0, 0, 0],
      mix_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_cycles=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_position_from_liquid_surface=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_surface_following_distance=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_speed=[0, 0, 0, 0, 0, 0, 0, 0],
      tube_section_height=[0, 0, 0, 0, 0, 0, 0, 0],
      tube_section_ratio=[0, 0, 0, 0, 0, 0, 0, 0],
      lld_mode=[0, 0, 0, 0, 0, 0, 0, 0],
      gamma_lld_sensitivity=[0, 0, 0, 0, 0, 0, 0, 0],
      dp_lld_sensitivity=[0, 0, 0, 0, 0, 0, 0, 0],
      lld_height_difference=[0, 0, 0, 0, 0, 0, 0, 0],
      tadm_enabled=False,
      limit_curve_index=[0, 0, 0, 0, 0, 0, 0, 0],
      recording_mode=0,
    )

    params = cmd.build_parameters()
    data = params.build()

    # Verify the parameters were built
    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Aspirate has 33 parameters
    self.assertEqual(len(results), 33)

    # Verify some key parameters
    # aspirate_type (i16_array)
    self.assertEqual(results[0][1], [0, 0, 0, 0, 0, 0, 0, 0])
    # channels_involved (u16_array)
    self.assertEqual(results[1][1], [1, 0, 0, 0, 0, 0, 0, 0])
    # aspirate_volume (u32_array) - at index 15
    self.assertEqual(results[15][1], [1000, 0, 0, 0, 0, 0, 0, 0])
    # tadm_enabled (bool) - at index 30
    self.assertEqual(results[30][1], False)
    # recording_mode (u16) - at index 32
    self.assertEqual(results[32][1], 0)

  def test_dispense_parameters(self):
    cmd = Dispense(
      dest=Address(1, 1, 257),
      dispense_type=[0, 0, 0, 0, 0, 0, 0, 0],
      channels_involved=[1, 0, 0, 0, 0, 0, 0, 0],
      x_positions=[10000, 0, 0, 0, 0, 0, 0, 0],
      y_positions=[20000, 0, 0, 0, 0, 0, 0, 0],
      minimum_traverse_height_at_beginning_of_a_command=14600,
      lld_search_height=[1000, 0, 0, 0, 0, 0, 0, 0],
      liquid_height=[500, 0, 0, 0, 0, 0, 0, 0],
      immersion_depth=[300, 0, 0, 0, 0, 0, 0, 0],
      surface_following_distance=[0, 0, 0, 0, 0, 0, 0, 0],
      minimum_height=[0, 0, 0, 0, 0, 0, 0, 0],
      min_z_endpos=0,
      swap_speed=[1000, 0, 0, 0, 0, 0, 0, 0],
      transport_air_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      dispense_volume=[1000, 0, 0, 0, 0, 0, 0, 0],  # 100uL in 0.1uL units
      stop_back_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      blow_out_air_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      dispense_speed=[500, 0, 0, 0, 0, 0, 0, 0],
      cut_off_speed=[0, 0, 0, 0, 0, 0, 0, 0],
      settling_time=[10, 0, 0, 0, 0, 0, 0, 0],
      mix_volume=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_cycles=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_position_from_liquid_surface=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_surface_following_distance=[0, 0, 0, 0, 0, 0, 0, 0],
      mix_speed=[0, 0, 0, 0, 0, 0, 0, 0],
      side_touch_off_distance=0,
      dispense_offset=[0, 0, 0, 0, 0, 0, 0, 0],
      tube_section_height=[0, 0, 0, 0, 0, 0, 0, 0],
      tube_section_ratio=[0, 0, 0, 0, 0, 0, 0, 0],
      lld_mode=[0, 0, 0, 0, 0, 0, 0, 0],
      gamma_lld_sensitivity=[0, 0, 0, 0, 0, 0, 0, 0],
      tadm_enabled=False,
      limit_curve_index=[0, 0, 0, 0, 0, 0, 0, 0],
      recording_mode=0,
    )

    params = cmd.build_parameters()
    data = params.build()

    parser = HoiParamsParser(data)
    results = parser.parse_all()

    # Dispense has 33 parameters
    self.assertEqual(len(results), 33)

    # Verify some key parameters
    # dispense_type (i16_array)
    self.assertEqual(results[0][1], [0, 0, 0, 0, 0, 0, 0, 0])
    # dispense_volume (u32_array)
    self.assertEqual(results[13][1], [1000, 0, 0, 0, 0, 0, 0, 0])


class TestNimbusBackendUnit(unittest.IsolatedAsyncioTestCase):
  """Unit tests for NimbusBackend class (no actual connection)."""

  async def test_backend_init(self):
    backend = NimbusBackend(host="192.168.1.100", port=2000)
    self.assertEqual(backend.io._host, "192.168.1.100")
    self.assertEqual(backend.io._port, 2000)
    self.assertIsNone(backend._num_channels)
    self.assertIsNone(backend._pipette_address)
    self.assertIsNone(backend._door_lock_address)
    self.assertIsNone(backend._nimbus_core_address)
    self.assertEqual(backend._channel_traversal_height, 146.0)

  async def test_backend_init_default_port(self):
    backend = NimbusBackend(host="192.168.1.100")
    self.assertEqual(backend.io._port, 2000)

  async def test_num_channels_before_setup_raises(self):
    backend = NimbusBackend(host="192.168.1.100")
    with self.assertRaises(RuntimeError) as ctx:
      _ = backend.num_channels
    self.assertIn("setup()", str(ctx.exception))

  async def test_set_minimum_channel_traversal_height(self):
    backend = NimbusBackend(host="192.168.1.100")
    backend.set_minimum_channel_traversal_height(100.0)
    self.assertEqual(backend._channel_traversal_height, 100.0)

  async def test_set_minimum_channel_traversal_height_invalid(self):
    backend = NimbusBackend(host="192.168.1.100")
    with self.assertRaises(ValueError):
      backend.set_minimum_channel_traversal_height(0)
    with self.assertRaises(ValueError):
      backend.set_minimum_channel_traversal_height(150)
    with self.assertRaises(ValueError):
      backend.set_minimum_channel_traversal_height(-10)

  async def test_fill_by_channels(self):
    backend = NimbusBackend(host="192.168.1.100")
    backend._num_channels = 8

    # Test with channels 0, 2, 4
    values = [100, 200, 300]
    use_channels = [0, 2, 4]
    result = backend._fill_by_channels(values, use_channels, default=0)

    expected = [100, 0, 200, 0, 300, 0, 0, 0]
    self.assertEqual(result, expected)

  async def test_fill_by_channels_mismatched_lengths(self):
    backend = NimbusBackend(host="192.168.1.100")
    backend._num_channels = 8

    with self.assertRaises(ValueError):
      backend._fill_by_channels([1, 2], [0, 1, 2], default=0)


def _mock_send_command_response(command) -> Optional[dict]:
  """Return appropriate mock responses based on command type."""
  if isinstance(command, IsTipPresent):
    return {"tip_present": [0] * 8}
  if isinstance(command, IsDoorLocked):
    return {"locked": True}
  if isinstance(command, IsInitialized):
    return {"initialized": True}
  if isinstance(command, GetChannelConfiguration_1):
    return {"channels": 8, "channel_types": [0] * 8}
  if isinstance(command, GetChannelConfiguration):
    return {"enabled": [False]}
  return None


def _setup_backend() -> NimbusBackend:
  """Create a NimbusBackend with pre-configured state for testing."""
  backend = NimbusBackend(host="192.168.1.100", port=2000)
  backend._num_channels = 8
  backend._pipette_address = Address(1, 1, 257)
  backend._door_lock_address = Address(1, 1, 268)
  backend._nimbus_core_address = Address(1, 1, 48896)
  backend._is_initialized = True
  return backend


def _setup_backend_with_deck(deck: NimbusDeck) -> NimbusBackend:
  """Create a NimbusBackend with pre-configured state and deck for testing."""
  backend = _setup_backend()
  backend._deck = deck
  return backend


class TestNimbusBackendCommands(unittest.IsolatedAsyncioTestCase):
  """Tests for NimbusBackend command methods."""

  async def asyncSetUp(self):
    self.backend = _setup_backend()
    self.mock_send = unittest.mock.AsyncMock(side_effect=_mock_send_command_response)
    self.backend.send_command = self.mock_send  # type: ignore[method-assign]

  def _get_command(self, cmd_type):
    for call in self.mock_send.call_args_list:
      if isinstance(call.args[0], cmd_type):
        return call.args[0]
    return None

  async def test_lock_door(self):
    await self.backend.lock_door()
    self.assertEqual(self.mock_send.call_count, 1)
    self.assertIsInstance(self._get_command(LockDoor), LockDoor)

  async def test_unlock_door(self):
    await self.backend.unlock_door()
    self.assertEqual(self.mock_send.call_count, 1)
    self.assertIsInstance(self._get_command(UnlockDoor), UnlockDoor)

  async def test_is_door_locked(self):
    result = await self.backend.is_door_locked()
    self.assertEqual(self.mock_send.call_count, 1)
    self.assertIsInstance(self._get_command(IsDoorLocked), IsDoorLocked)
    self.assertTrue(result)

  async def test_park(self):
    await self.backend.park()
    self.assertEqual(self.mock_send.call_count, 1)
    self.assertIsInstance(self._get_command(Park), Park)

  async def test_door_methods_without_address_raise(self):
    self.backend._door_lock_address = None

    with self.assertRaises(RuntimeError):
      await self.backend.lock_door()

    with self.assertRaises(RuntimeError):
      await self.backend.unlock_door()

    with self.assertRaises(RuntimeError):
      await self.backend.is_door_locked()

  async def test_park_without_address_raises(self):
    self.backend._nimbus_core_address = None

    with self.assertRaises(RuntimeError):
      await self.backend.park()


class TestNimbusBackendSerialization(unittest.IsolatedAsyncioTestCase):
  """Tests for NimbusBackend serialization."""

  async def test_serialize(self):
    backend = NimbusBackend(host="192.168.1.100", port=2000)
    backend._client_id = 5

    serialized = backend.serialize()
    self.assertEqual(serialized["client_id"], 5)
    self.assertIn("instrument_addresses", serialized)


class TestNimbusLiquidHandling(unittest.IsolatedAsyncioTestCase):
  """Tests for NimbusBackend liquid handling command generation."""

  async def asyncSetUp(self):
    self.deck = NimbusDeck()
    self.backend = _setup_backend_with_deck(self.deck)
    self.mock_send = unittest.mock.AsyncMock(side_effect=_mock_send_command_response)
    self.backend.send_command = self.mock_send  # type: ignore[method-assign]

    self.tip_rack = hamilton_96_tiprack_300uL("tip_rack")
    self.deck.assign_child_resource(self.tip_rack, rails=1)

    self.plate = Cor_96_wellplate_360ul_Fb("plate")
    self.deck.assign_child_resource(self.plate, rails=10)

    self.tip = HamiltonTip(
      name="test_tip",
      has_filter=False,
      total_tip_length=59.9,
      maximal_volume=300.0,
      tip_size=TipSize.STANDARD_VOLUME,
      pickup_method=TipPickupMethod.OUT_OF_RACK,
    )

  def _get_commands(self, cmd_type):
    return [
      call.args[0] for call in self.mock_send.call_args_list if isinstance(call.args[0], cmd_type)
    ]

  # === Pick up tips tests ===

  async def test_pick_up_tips_single_channel(self):
    tip_spot = self.tip_rack.get_item("A1")
    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip_spot.get_tip())],
      use_channels=[0],
    )

    cmds = self._get_commands(PickupTips)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])
    self.assertNotEqual(cmds[0].x_positions[0], 0)
    self.assertNotEqual(cmds[0].y_positions[0], 0)
    self.assertEqual(cmds[0].x_positions[1], 0)
    self.assertEqual(cmds[0].tip_types[0], NimbusTipType.STANDARD_300UL)

  async def test_pick_up_tips_multiple_channels(self):
    tip_spot_0 = self.tip_rack.get_item("A1")
    tip_spot_1 = self.tip_rack.get_item("B1")
    ops = [
      Pickup(resource=tip_spot_0, offset=Coordinate.zero(), tip=tip_spot_0.get_tip()),
      Pickup(resource=tip_spot_1, offset=Coordinate.zero(), tip=tip_spot_1.get_tip()),
    ]

    await self.backend.pick_up_tips(ops, use_channels=[0, 2])

    cmds = self._get_commands(PickupTips)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 1, 0, 0, 0, 0, 0])

  async def test_pick_up_tips_traversal_height(self):
    tip_spot = self.tip_rack.get_item("A1")
    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip_spot.get_tip())],
      use_channels=[0],
      minimum_traverse_height_at_beginning_of_a_command=100.0,
    )

    cmds = self._get_commands(PickupTips)
    self.assertEqual(cmds[0].minimum_traverse_height_at_beginning_of_a_command, 10000)

  # === Drop tips tests ===

  async def test_drop_tips_to_waste_uses_roll(self):
    waste_pos = self.deck.get_resource("default_long_1")
    await self.backend.drop_tips(
      [Drop(resource=waste_pos, offset=Coordinate.zero(), tip=self.tip)], use_channels=[0]
    )

    cmds = self._get_commands(DropTipsRoll)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])
    self.assertNotEqual(cmds[0].roll_distances[0], 0)

  async def test_drop_tips_to_rack_uses_drop_tips(self):
    tip_spot = self.tip_rack.get_item("A1")
    await self.backend.drop_tips(
      [Drop(resource=tip_spot, offset=Coordinate.zero(), tip=self.tip)], use_channels=[0]
    )

    cmds = self._get_commands(DropTips)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])
    self.assertFalse(cmds[0].default_waste)

  async def test_drop_tips_multiple_channels_to_waste(self):
    ops = [
      Drop(
        resource=self.deck.get_resource("default_long_1"), offset=Coordinate.zero(), tip=self.tip
      ),
      Drop(
        resource=self.deck.get_resource("default_long_3"), offset=Coordinate.zero(), tip=self.tip
      ),
    ]
    await self.backend.drop_tips(ops, use_channels=[0, 2])

    cmds = self._get_commands(DropTipsRoll)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 1, 0, 0, 0, 0, 0])

  # === Aspirate tests ===

  async def test_aspirate_single_channel(self):
    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=5.0,
          blow_out_air_volume=40.0,
          mix=None,
        )
      ],
      use_channels=[0],
    )

    cmds = self._get_commands(Aspirate)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])
    self.assertEqual(cmds[0].aspirate_volume[0], 1000)
    self.assertEqual(cmds[0].aspiration_speed[0], 500)
    self.assertEqual(cmds[0].blow_out_air_volume[0], 400)

  async def test_aspirate_default_flow_rate(self):
    """Test that flow_rate=None uses tip-based default (100 uL/s for 300ul tip)."""
    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=None,
          liquid_height=5.0,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
    )

    cmds = self._get_commands(Aspirate)
    self.assertEqual(cmds[0].aspiration_speed[0], 1000)  # 100 uL/s * 10

  async def test_aspirate_multiple_channels(self):
    ops = [
      SingleChannelAspiration(
        resource=self.plate.get_item("A1"),
        offset=Coordinate.zero(),
        tip=self.tip,
        volume=100.0,
        flow_rate=50.0,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      ),
      SingleChannelAspiration(
        resource=self.plate.get_item("A2"),
        offset=Coordinate.zero(),
        tip=self.tip,
        volume=150.0,
        flow_rate=75.0,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      ),
    ]
    await self.backend.aspirate(ops, use_channels=[1, 3])

    cmds = self._get_commands(Aspirate)
    self.assertEqual(cmds[0].channels_involved, [0, 1, 0, 1, 0, 0, 0, 0])
    self.assertEqual(cmds[0].aspirate_volume[1], 1000)
    self.assertEqual(cmds[0].aspirate_volume[3], 1500)
    self.assertEqual(cmds[0].aspiration_speed[1], 500)
    self.assertEqual(cmds[0].aspiration_speed[3], 750)

  async def test_aspirate_with_adc_enabled(self):
    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
      adc_enabled=True,
    )

    cmds = self._get_commands(EnableADC)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])

  async def test_aspirate_with_adc_disabled(self):
    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
      adc_enabled=False,
    )

    cmds = self._get_commands(DisableADC)
    self.assertEqual(len(cmds), 1)

  # === Dispense tests ===

  async def test_dispense_single_channel(self):
    await self.backend.dispense(
      [
        SingleChannelDispense(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=5.0,
          blow_out_air_volume=40.0,
          mix=None,
        )
      ],
      use_channels=[0],
    )

    cmds = self._get_commands(Dispense)
    self.assertEqual(len(cmds), 1)
    self.assertEqual(cmds[0].channels_involved, [1, 0, 0, 0, 0, 0, 0, 0])
    self.assertEqual(cmds[0].dispense_volume[0], 1000)
    self.assertEqual(cmds[0].dispense_speed[0], 500)
    self.assertEqual(cmds[0].blow_out_air_volume[0], 400)

  async def test_dispense_multiple_channels(self):
    ops = [
      SingleChannelDispense(
        resource=self.plate.get_item("A1"),
        offset=Coordinate.zero(),
        tip=self.tip,
        volume=100.0,
        flow_rate=50.0,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      ),
      SingleChannelDispense(
        resource=self.plate.get_item("A2"),
        offset=Coordinate.zero(),
        tip=self.tip,
        volume=200.0,
        flow_rate=100.0,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      ),
    ]
    await self.backend.dispense(ops, use_channels=[2, 5])

    cmds = self._get_commands(Dispense)
    self.assertEqual(cmds[0].channels_involved, [0, 0, 1, 0, 0, 1, 0, 0])
    self.assertEqual(cmds[0].dispense_volume[2], 1000)
    self.assertEqual(cmds[0].dispense_volume[5], 2000)

  async def test_dispense_with_adc_enabled(self):
    await self.backend.dispense(
      [
        SingleChannelDispense(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
      adc_enabled=True,
    )

    cmds = self._get_commands(EnableADC)
    self.assertEqual(len(cmds), 1)

  # === Coordinate conversion tests ===

  async def test_aspirate_coordinates(self):
    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=self.plate.get_item("A1"),
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
    )

    cmds = self._get_commands(Aspirate)
    self.assertEqual(cmds[0].x_positions[0], 8726)
    self.assertEqual(cmds[0].y_positions[0], -28972)

  async def test_offset_applied_to_coordinates(self):
    well = self.plate.get_item("A1")

    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=well,
          offset=Coordinate.zero(),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
    )
    x_no_offset = self._get_commands(Aspirate)[0].x_positions[0]

    self.mock_send.reset_mock()

    await self.backend.aspirate(
      [
        SingleChannelAspiration(
          resource=well,
          offset=Coordinate(x=10.0, y=0.0, z=0.0),
          tip=self.tip,
          volume=100.0,
          flow_rate=50.0,
          liquid_height=None,
          blow_out_air_volume=None,
          mix=None,
        )
      ],
      use_channels=[0],
    )
    x_with_offset = self._get_commands(Aspirate)[0].x_positions[0]

    self.assertEqual(x_with_offset - x_no_offset, 1000)


class TestNimbusTipPickupDropAllSizes(unittest.IsolatedAsyncioTestCase):
  """Tests for Nimbus tip pickup/drop Z positions across all tip sizes.

  These tests verify that the begin/end tip pickup and drop process values
  match the machine-validated values.
  """

  async def asyncSetUp(self):
    self.deck = NimbusDeck()
    self.backend = _setup_backend_with_deck(self.deck)
    self.mock_send = unittest.mock.AsyncMock(side_effect=_mock_send_command_response)
    self.backend.send_command = self.mock_send  # type: ignore[method-assign]

  def _get_commands(self, cmd_type):
    return [
      call.args[0] for call in self.mock_send.call_args_list if isinstance(call.args[0], cmd_type)
    ]

  async def test_10uL_tips(self):
    from pylabrobot.resources.hamilton.tip_racks import hamilton_96_tiprack_10uL

    tip_rack = hamilton_96_tiprack_10uL("tips")
    self.deck.assign_child_resource(tip_rack, rails=1)
    tip_spot = tip_rack.get_item("A1")
    tip = tip_spot.get_tip()

    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    pickup_cmd = self._get_commands(PickupTips)[0]
    self.assertEqual(pickup_cmd.begin_tip_pick_up_process[0], 740)
    self.assertEqual(pickup_cmd.end_tip_pick_up_process[0], -60)

    self.mock_send.reset_mock()
    await self.backend.drop_tips(
      [Drop(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    drop_cmd = self._get_commands(DropTips)[0]
    self.assertEqual(drop_cmd.begin_tip_deposit_process[0], -1250)
    self.assertEqual(drop_cmd.end_tip_deposit_process[0], -2250)

    tip_rack.unassign()

  async def test_50uL_tips(self):
    from pylabrobot.resources.hamilton.tip_racks import hamilton_96_tiprack_50uL

    tip_rack = hamilton_96_tiprack_50uL("tips")
    self.deck.assign_child_resource(tip_rack, rails=1)
    tip_spot = tip_rack.get_item("A1")
    tip = tip_spot.get_tip()

    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    pickup_cmd = self._get_commands(PickupTips)[0]
    self.assertEqual(pickup_cmd.begin_tip_pick_up_process[0], 990)
    self.assertEqual(pickup_cmd.end_tip_pick_up_process[0], 190)

    self.mock_send.reset_mock()
    await self.backend.drop_tips(
      [Drop(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    drop_cmd = self._get_commands(DropTips)[0]
    self.assertEqual(drop_cmd.begin_tip_deposit_process[0], -3050)
    self.assertEqual(drop_cmd.end_tip_deposit_process[0], -4050)

    tip_rack.unassign()

  async def test_300uL_tips(self):
    from pylabrobot.resources.hamilton.tip_racks import hamilton_96_tiprack_300uL

    tip_rack = hamilton_96_tiprack_300uL("tips")
    self.deck.assign_child_resource(tip_rack, rails=1)
    tip_spot = tip_rack.get_item("A1")
    tip = tip_spot.get_tip()

    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    pickup_cmd = self._get_commands(PickupTips)[0]
    self.assertEqual(pickup_cmd.begin_tip_pick_up_process[0], 940)
    self.assertEqual(pickup_cmd.end_tip_pick_up_process[0], 140)

    self.mock_send.reset_mock()
    await self.backend.drop_tips(
      [Drop(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    drop_cmd = self._get_commands(DropTips)[0]
    self.assertEqual(drop_cmd.begin_tip_deposit_process[0], -4050)
    self.assertEqual(drop_cmd.end_tip_deposit_process[0], -5050)

    tip_rack.unassign()

  async def test_1000uL_tips(self):
    from pylabrobot.resources.hamilton.tip_racks import hamilton_96_tiprack_1000uL

    tip_rack = hamilton_96_tiprack_1000uL("tips")
    self.deck.assign_child_resource(tip_rack, rails=1)
    tip_spot = tip_rack.get_item("A1")
    tip = tip_spot.get_tip()

    await self.backend.pick_up_tips(
      [Pickup(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    pickup_cmd = self._get_commands(PickupTips)[0]
    self.assertEqual(pickup_cmd.begin_tip_pick_up_process[0], 1160)
    self.assertEqual(pickup_cmd.end_tip_pick_up_process[0], 360)

    self.mock_send.reset_mock()
    await self.backend.drop_tips(
      [Drop(resource=tip_spot, offset=Coordinate.zero(), tip=tip)],
      use_channels=[0],
    )
    drop_cmd = self._get_commands(DropTips)[0]
    self.assertEqual(drop_cmd.begin_tip_deposit_process[0], -7350)
    self.assertEqual(drop_cmd.end_tip_deposit_process[0], -8350)

    tip_rack.unassign()


if __name__ == "__main__":
  unittest.main()



from pylabrobot.resources.plate import Plate
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.well import (
  CrossSectionType,
  Well,
  WellBottomType,
)


def BioRad_384_wellplate_50uL_Vb(name: str) -> Plate:
  return Plate(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=10.40,
    lid=None,
    model="BioRad_384_wellplate_50uL_Vb",
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=24,
      num_items_y=16,
      dx=10.58,
      dy=7.44,
      dz=1.05,
      item_dx=4.5,
      item_dy=4.5,
      size_x=3.10,
      size_y=3.10,
      size_z=9.35,
      bottom_type=WellBottomType.V,
      material_z_thickness=1,  # measured
      cross_section_type=CrossSectionType.CIRCLE,
    ),
  )



from typing import Dict


class NoChannelError(Exception):
  """Raised when no channel is available, e.g. when trying to pick up a tip with no empty channels.
  This error is only raised when the channel is automatically selected by the system.

  Examples:
  - when trying to pick up a tip with no empty channels available on a robot
  """


class ChannelizedError(Exception):
  """Raised by operations that work on multiple channels: pick_up_tips, drop_tips, aspirate, and
  dispense. Contains a key for each channel that had an error, and the error that occurred."""

  def __init__(self, errors: Dict[int, Exception], **kwargs):
    self.errors = errors
    self.kwargs = kwargs

  def __str__(self) -> str:
    kwarg_string = ", ".join([f"{k}={v}" for k, v in self.kwargs.items()])
    return f"ChannelizedError(errors={self.errors}, {kwarg_string})"

  def __len__(self) -> int:
    return len(self.errors)



"""Tecan Infinite 200 PRO backend.

This backend targets the Infinite "M" series (e.g., Infinite 200 PRO).  The
"F" series uses a different optical path and is not covered here.
"""

from __future__ import annotations

import asyncio
import logging
import math
import re
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple

from pylabrobot.io.binary import Reader
from pylabrobot.io.usb import USB
from pylabrobot.plate_reading.backend import PlateReaderBackend
from pylabrobot.resources import Plate
from pylabrobot.resources.well import Well

logger = logging.getLogger(__name__)
BIN_RE = re.compile(r"^(\d+),BIN:$")


def _integration_microseconds_to_seconds(value: int) -> float:
  # DLL/UI indicates integration time is stored in microseconds; UI displays ms by dividing by 1000.
  return value / 1_000_000.0


def _is_abs_calibration_len(payload_len: int) -> bool:
  return payload_len >= 22 and (payload_len - 4) % 18 == 0


def _is_abs_data_len(payload_len: int) -> bool:
  return payload_len >= 14 and (payload_len - 4) % 10 == 0


def _split_payload_and_trailer(
  payload_len: int, blob: bytes
) -> Optional[Tuple[bytes, Tuple[int, int]]]:
  if len(blob) != payload_len + 4:
    return None
  payload = blob[:payload_len]
  trailer_reader = Reader(blob[payload_len:], little_endian=False)
  return payload, (trailer_reader.u16(), trailer_reader.u16())


@dataclass(frozen=True)
class _AbsorbanceCalibrationItem:
  ticker_overflows: int
  ticker_counter: int
  meas_gain: int
  meas_dark: int
  meas_bright: int
  ref_gain: int
  ref_dark: int
  ref_bright: int


@dataclass(frozen=True)
class _AbsorbanceCalibration:
  ex: int
  items: List[_AbsorbanceCalibrationItem]


def _decode_abs_calibration(payload_len: int, blob: bytes) -> Optional[_AbsorbanceCalibration]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) < 4 + 18:
    return None
  if (len(payload) - 4) % 18 != 0:
    return None
  reader = Reader(payload, little_endian=False)
  reader.raw_bytes(2)  # skip first 2 bytes
  ex = reader.u16()
  items: List[_AbsorbanceCalibrationItem] = []
  while reader.has_remaining():
    items.append(
      _AbsorbanceCalibrationItem(
        ticker_overflows=reader.u32(),
        ticker_counter=reader.u16(),
        meas_gain=reader.u16(),
        meas_dark=reader.u16(),
        meas_bright=reader.u16(),
        ref_gain=reader.u16(),
        ref_dark=reader.u16(),
        ref_bright=reader.u16(),
      )
    )
  return _AbsorbanceCalibration(ex=ex, items=items)


def _decode_abs_data(
  payload_len: int, blob: bytes
) -> Optional[Tuple[int, int, List[Tuple[int, int]]]]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) < 4:
    return None
  reader = Reader(payload, little_endian=False)
  label = reader.u16()
  ex = reader.u16()
  items: List[Tuple[int, int]] = []
  while reader.offset() + 10 <= len(payload):
    reader.raw_bytes(6)  # skip first 6 bytes of each item
    meas = reader.u16()
    ref = reader.u16()
    items.append((meas, ref))
  if reader.offset() != len(payload):
    return None
  return label, ex, items


def _absorbance_od_calibrated(
  cal: _AbsorbanceCalibration, meas_ref_items: List[Tuple[int, int]], od_max: float = 4.0
) -> float:
  if not cal.items:
    raise ValueError("ABS calibration packet contained no calibration items.")

  min_corr_trans = math.pow(10.0, -od_max)

  if len(cal.items) == len(meas_ref_items) and len(cal.items) > 1:
    corr_trans_vals: List[float] = []
    for (meas, ref), cal_item in zip(meas_ref_items, cal.items):
      denom_corr = cal_item.meas_bright - cal_item.meas_dark
      if denom_corr == 0:
        continue
      f_corr = (cal_item.ref_bright - cal_item.ref_dark) / denom_corr
      denom = ref - cal_item.ref_dark
      if denom == 0:
        continue
      corr_trans_vals.append(((meas - cal_item.meas_dark) / denom) * f_corr)
    if not corr_trans_vals:
      raise ZeroDivisionError("ABS invalid: no usable reads after per-read calibration.")
    corr_trans = max(sum(corr_trans_vals) / len(corr_trans_vals), min_corr_trans)
    return float(-math.log10(corr_trans))

  cal0 = cal.items[0]
  denom_corr = cal0.meas_bright - cal0.meas_dark
  if denom_corr == 0:
    raise ZeroDivisionError("ABS calibration invalid: meas_bright == meas_dark")
  f_corr = (cal0.ref_bright - cal0.ref_dark) / denom_corr

  trans_vals: List[float] = []
  for meas, ref in meas_ref_items:
    denom = ref - cal0.ref_dark
    if denom == 0:
      continue
    trans_vals.append((meas - cal0.meas_dark) / denom)
  if not trans_vals:
    raise ZeroDivisionError("ABS invalid: all ref reads equal ref_dark")

  trans_mean = sum(trans_vals) / len(trans_vals)
  corr_trans = max(trans_mean * f_corr, min_corr_trans)
  return float(-math.log10(corr_trans))


@dataclass(frozen=True)
class _FluorescenceCalibration:
  ex: int
  meas_dark: int
  ref_dark: int
  ref_bright: int


def _decode_flr_calibration(payload_len: int, blob: bytes) -> Optional[_FluorescenceCalibration]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) != 18:
    return None
  reader = Reader(payload, little_endian=False)
  ex = reader.u16()
  reader.raw_bytes(8)  # skip bytes 2-9
  meas_dark = reader.u16()
  reader.raw_bytes(2)  # skip bytes 12-13
  ref_dark = reader.u16()
  ref_bright = reader.u16()
  return _FluorescenceCalibration(
    ex=ex,
    meas_dark=meas_dark,
    ref_dark=ref_dark,
    ref_bright=ref_bright,
  )


def _decode_flr_data(
  payload_len: int, blob: bytes
) -> Optional[Tuple[int, int, int, List[Tuple[int, int]]]]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) < 6:
    return None
  reader = Reader(payload, little_endian=False)
  label = reader.u16()
  ex = reader.u16()
  em = reader.u16()
  items: List[Tuple[int, int]] = []
  while reader.offset() + 10 <= len(payload):
    reader.raw_bytes(6)  # skip first 6 bytes of each item
    meas = reader.u16()
    ref = reader.u16()
    items.append((meas, ref))
  if reader.offset() != len(payload):
    return None
  return label, ex, em, items


def _fluorescence_corrected(
  cal: _FluorescenceCalibration, meas_ref_items: List[Tuple[int, int]]
) -> int:
  if not meas_ref_items:
    return 0
  meas_mean = sum(m for m, _ in meas_ref_items) / len(meas_ref_items)
  ref_mean = sum(r for _, r in meas_ref_items) / len(meas_ref_items)
  denom = ref_mean - cal.ref_dark
  if denom == 0:
    return 0
  corr = (meas_mean - cal.meas_dark) * (cal.ref_bright - cal.ref_dark) / denom
  return int(round(corr))


@dataclass(frozen=True)
class _LuminescenceCalibration:
  ref_dark: int


def _decode_lum_calibration(payload_len: int, blob: bytes) -> Optional[_LuminescenceCalibration]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) != 10:
    return None
  reader = Reader(payload, little_endian=False)
  reader.raw_bytes(6)  # skip bytes 0-5
  return _LuminescenceCalibration(ref_dark=reader.i32())


def _decode_lum_data(payload_len: int, blob: bytes) -> Optional[Tuple[int, int, List[int]]]:
  split = _split_payload_and_trailer(payload_len, blob)
  if split is None:
    return None
  payload, _ = split
  if len(payload) < 4:
    return None
  reader = Reader(payload, little_endian=False)
  label = reader.u16()
  em = reader.u16()
  counts: List[int] = []
  while reader.offset() + 10 <= len(payload):
    reader.raw_bytes(6)  # skip first 6 bytes of each item
    counts.append(reader.i32())
  if reader.offset() != len(payload):
    return None
  return label, em, counts


def _luminescence_intensity(
  cal: _LuminescenceCalibration,
  counts: List[int],
  dark_integration_s: float,
  meas_integration_s: float,
) -> int:
  if not counts:
    return 0
  if dark_integration_s == 0 or meas_integration_s == 0:
    return 0
  count_mean = sum(counts) / len(counts)
  corrected_rate = (count_mean / meas_integration_s) - (cal.ref_dark / dark_integration_s)
  return int(corrected_rate)


StagePosition = Tuple[int, int]


def _consume_leading_ascii_frame(buffer: bytearray) -> Tuple[bool, Optional[str]]:
  """Remove a leading STX...ETX ASCII frame if present."""

  if not buffer or buffer[0] != 0x02:
    return False, None
  end = buffer.find(b"\x03", 1)
  if end == -1:
    return False, None
  # Payload is followed by a 4-byte trailer and optional CR.
  if len(buffer) < end + 5:
    return False, None
  text = buffer[1:end].decode("ascii", "ignore")
  del buffer[: end + 5]
  if buffer and buffer[0] == 0x0D:
    del buffer[0]
  return True, text


def _consume_status_frame(buffer: bytearray, length: int) -> bool:
  """Drop a leading ESC-prefixed status frame if present."""

  if len(buffer) >= length and buffer[0] == 0x1B:
    del buffer[:length]
    return True
  return False


@dataclass
class _StreamEvent:
  """Parsed stream event (ASCII or binary)."""

  text: Optional[str] = None
  payload_len: Optional[int] = None
  blob: Optional[bytes] = None


class _StreamParser:
  """Parse mixed ASCII and binary packets from the reader."""

  def __init__(
    self,
    *,
    status_frame_len: Optional[int] = None,
    allow_bare_ascii: bool = False,
  ) -> None:
    """Initialize the stream parser."""
    self._buffer = bytearray()
    self._pending_bin: Optional[int] = None
    self._status_frame_len = status_frame_len
    self._allow_bare_ascii = allow_bare_ascii

  def has_pending_bin(self) -> bool:
    """Return True if a binary payload length is pending."""
    return self._pending_bin is not None

  def feed(self, chunk: bytes) -> List[_StreamEvent]:
    """Feed raw bytes and return newly parsed events."""
    self._buffer.extend(chunk)
    events: List[_StreamEvent] = []
    progressed = True
    while progressed:
      progressed = False
      if self._pending_bin is not None:
        need = self._pending_bin + 4
        if len(self._buffer) < need:
          break
        blob = bytes(self._buffer[:need])
        del self._buffer[:need]
        events.append(_StreamEvent(payload_len=self._pending_bin, blob=blob))
        self._pending_bin = None
        progressed = True
        continue
      if self._status_frame_len and _consume_status_frame(self._buffer, self._status_frame_len):
        progressed = True
        continue
      consumed, text = _consume_leading_ascii_frame(self._buffer)
      if consumed:
        events.append(_StreamEvent(text=text))
        if text:
          m = BIN_RE.match(text)
          if m:
            self._pending_bin = int(m.group(1))
        progressed = True
        continue
      if self._allow_bare_ascii and self._buffer and all(32 <= b <= 126 for b in self._buffer):
        text = self._buffer.decode("ascii", "ignore")
        self._buffer.clear()
        events.append(_StreamEvent(text=text))
        progressed = True
        continue
    return events


class _MeasurementDecoder(ABC):
  """Shared incremental decoder for Infinite measurement streams."""

  STATUS_FRAME_LEN: Optional[int] = None

  def __init__(self, expected: int) -> None:
    """Initialize decoder state for a scan with expected measurements."""
    self.expected = expected
    self._terminal_seen = False
    self._parser = _StreamParser(status_frame_len=self.STATUS_FRAME_LEN)

  @property
  @abstractmethod
  def count(self) -> int:
    """Return number of decoded measurements so far."""

  @property
  def done(self) -> bool:
    """Return True if the decoder has seen all expected measurements."""
    return self.count >= self.expected

  def pop_terminal(self) -> bool:
    """Return and clear the terminal frame seen flag."""
    seen = self._terminal_seen
    self._terminal_seen = False
    return seen

  def feed(self, chunk: bytes) -> None:
    """Consume a raw chunk and update decoder state."""
    for event in self._parser.feed(chunk):
      if event.text is not None:
        if event.text == "ST":
          self._terminal_seen = True
      elif event.payload_len is not None and event.blob is not None:
        self.feed_bin(event.payload_len, event.blob)

  def feed_bin(self, payload_len: int, blob: bytes) -> None:
    """Handle a binary payload if the decoder expects one."""
    if self._should_consume_bin(payload_len):
      self._handle_bin(payload_len, blob)

  def _should_consume_bin(self, _payload_len: int) -> bool:
    return False

  def _handle_bin(self, _payload_len: int, _blob: bytes) -> None:
    return None


class ExperimentalTecanInfinite200ProBackend(PlateReaderBackend):
  """Backend shell for the Infinite 200 PRO."""

  _MODE_CAPABILITY_COMMANDS: Dict[str, List[str]] = {
    "ABS": [
      "#BEAM DIAMETER",
      # Additional capabilities available but currently unused:
      # "#EXCITATION WAVELENGTH",
      # "#EXCITATION USAGE",
      # "#EXCITATION NAME",
      # "#EXCITATION BANDWIDTH",
      # "#EXCITATION ATTENUATION",
      # "#EXCITATION DESCRIPTION",
      # "#TIME READDELAY",
      # "#SHAKING MODE",
      # "#SHAKING CONST.ORBITAL",
      # "#SHAKING AMPLITUDE",
      # "#SHAKING TIME",
      # "#SHAKING CONST.LINEAR",
      # "#TEMPERATURE PLATE",
    ],
    "FI.TOP": [
      # "#BEAM DIAMETER",
      # Additional capabilities available but currently unused:
      # "#EMISSION WAVELENGTH",
      # "#EMISSION USAGE",
      # "#EMISSION NAME",
      # "#EMISSION BANDWIDTH",
      # "#EMISSION ATTENUATION",
      # "#EMISSION DESCRIPTION",
      # "#EXCITATION WAVELENGTH",
      # "#EXCITATION USAGE",
      # "#EXCITATION NAME",
      # "#EXCITATION BANDWIDTH",
      # "#EXCITATION ATTENUATION",
      # "#EXCITATION DESCRIPTION",
      # "#TIME INTEGRATION",
      # "#TIME LAG",
      # "#TIME READDELAY",
      # "#GAIN VALUE",
      # "#READS SPEED",
      # "#READS NUMBER",
      # "#RANGES PMT,EXCITATION",
      # "#RANGES PMT,EMISSION",
      # "#POSITION FIL,Z",
      # "#TEMPERATURE PLATE",
    ],
    "FI.BOTTOM": [
      # "#BEAM DIAMETER",
      # Additional capabilities available but currently unused:
      # "#EMISSION WAVELENGTH",
      # "#EMISSION USAGE",
      # "#EXCITATION WAVELENGTH",
      # "#EXCITATION USAGE",
      # "#TIME INTEGRATION",
      # "#TIME LAG",
      # "#TIME READDELAY",
    ],
    "LUM": [
      # "#BEAM DIAMETER",
      # Additional capabilities available but currently unused:
      # "#EMISSION WAVELENGTH",
      # "#EMISSION USAGE",
      # "#EMISSION NAME",
      # "#EMISSION BANDWIDTH",
      # "#EMISSION ATTENUATION",
      # "#EMISSION DESCRIPTION",
      # "#TIME INTEGRATION",
      # "#TIME READDELAY",
    ],
  }

  VENDOR_ID = 0x0C47
  PRODUCT_ID = 0x8007

  def __init__(
    self,
    counts_per_mm_x: float = 1_000,
    counts_per_mm_y: float = 1_000,
    counts_per_mm_z: float = 1_000,
  ) -> None:
    super().__init__()
    self.io = USB(
      id_vendor=self.VENDOR_ID,
      id_product=self.PRODUCT_ID,
      human_readable_device_name="Tecan Infinite 200 PRO",
      packet_read_timeout=3,
      read_timeout=30,
    )
    self.counts_per_mm_x = counts_per_mm_x
    self.counts_per_mm_y = counts_per_mm_y
    self.counts_per_mm_z = counts_per_mm_z
    self._setup_lock = asyncio.Lock()
    self._ready = False
    self._read_chunk_size = 512
    self._max_row_wait_s = 300.0
    self._mode_capabilities: Dict[str, Dict[str, str]] = {}
    self._pending_bin_events: List[Tuple[int, bytes]] = []
    self._parser = _StreamParser(allow_bare_ascii=True)
    self._run_active = False
    self._active_step_loss_commands: List[str] = []

  async def setup(self) -> None:
    async with self._setup_lock:
      if self._ready:
        return
      await self.io.setup()
      await self._initialize_device()
      for mode in self._MODE_CAPABILITY_COMMANDS:
        if mode not in self._mode_capabilities:
          await self._query_mode_capabilities(mode)
      self._ready = True

  async def stop(self) -> None:
    async with self._setup_lock:
      if not self._ready:
        return
      await self._cleanup_protocol()
      await self.io.stop()
      self._mode_capabilities.clear()
      self._reset_stream_state()
      self._ready = False

  async def open(self) -> None:
    """Open the reader drawer."""

    await self._send_command("ABSOLUTE MTP,OUT")
    await self._send_command("BY#T5000")

  async def close(self, plate: Optional[Plate]) -> None:  # noqa: ARG002
    """Close the reader drawer."""

    await self._send_command("ABSOLUTE MTP,IN")
    await self._send_command("BY#T5000")

  async def _run_scan(
    self,
    ordered_wells: Sequence[Well],
    decoder: _MeasurementDecoder,
    mode: str,
    step_loss_commands: List[str],
    serpentine: bool,
    scan_direction: str,
  ) -> None:
    """Run the common scan loop for all measurement types.

    Args:
      ordered_wells: The wells to scan in row-major order.
      decoder: The decoder to use for parsing measurements.
      mode: The mode name for logging (e.g., "Absorbance").
      step_loss_commands: Commands to run after the scan to check for step loss.
      serpentine: Whether to use serpentine scan order.
      scan_direction: The scan direction command (e.g., "ALTUP", "UP").
    """
    self._active_step_loss_commands = step_loss_commands

    for row_index, row_wells in self._group_by_row(ordered_wells):
      start_x, end_x, count = self._scan_range(row_index, row_wells, serpentine=serpentine)
      _, y_stage = self._map_well_to_stage(row_wells[0])

      await self._send_command(f"ABSOLUTE MTP,Y={y_stage}")
      await self._send_command(f"SCAN DIRECTION={scan_direction}")
      await self._send_command(
        f"SCANX {start_x},{end_x},{count}", wait_for_terminal=False, read_response=False
      )
      logger.info(
        "Queued %s scan row %s (%s wells): y=%s, x=%s..%s",
        mode.lower(),
        row_index,
        count,
        y_stage,
        start_x,
        end_x,
      )
      await self._await_measurements(decoder, count, mode)
      await self._await_scan_terminal(decoder.pop_terminal())

  async def read_absorbance(
    self,
    plate: Plate,
    wells: List[Well],
    wavelength: int,
    flashes: int = 25,
    bandwidth: Optional[float] = None,
  ) -> List[Dict]:
    """Queue and execute an absorbance scan.

    Args:
      bandwidth: Excitation bandwidth in nm. If None, auto-selected (9 nm for >315 nm, 5 nm
        otherwise).
    """

    if not 230 <= wavelength <= 1_000:
      raise ValueError("Absorbance wavelength must be between 230 nm and 1000 nm.")

    ordered_wells = wells if wells else plate.get_all_items()
    scan_wells = self._scan_visit_order(ordered_wells, serpentine=True)
    decoder = _AbsorbanceRunDecoder(len(scan_wells))

    await self._begin_run()
    try:
      await self._configure_absorbance(wavelength, flashes=flashes, bandwidth=bandwidth)
      await self._run_scan(
        ordered_wells=ordered_wells,
        decoder=decoder,
        mode="Absorbance",
        step_loss_commands=["CHECK MTP.STEPLOSS", "CHECK ABS.STEPLOSS"],
        serpentine=True,
        scan_direction="ALTUP",
      )

      self._drain_pending_bin_events(decoder)
      if len(decoder.measurements) != len(scan_wells):
        raise RuntimeError("Absorbance decoder did not complete scan.")
      intensities: List[float] = []
      cal = decoder.calibration
      if cal is None:
        raise RuntimeError("ABS calibration packet not seen; cannot compute calibrated OD.")
      for meas in decoder.measurements:
        items = meas.items or [(meas.sample, meas.reference)]
        od = _absorbance_od_calibrated(cal, items)
        intensities.append(od)
      matrix = self._format_plate_result(plate, scan_wells, intensities)
      return [
        {
          "wavelength": wavelength,
          "time": time.time(),
          "temperature": None,
          "data": matrix,
        }
      ]
    finally:
      await self._end_run()

  async def _clear_mode_settings(self, excitation: bool = False, emission: bool = False) -> None:
    """Clear mode settings before configuring a new scan."""
    if excitation:
      await self._send_command("EXCITATION CLEAR", allow_timeout=True)
    if emission:
      await self._send_command("EMISSION CLEAR", allow_timeout=True)
    await self._send_command("TIME CLEAR", allow_timeout=True)
    await self._send_command("GAIN CLEAR", allow_timeout=True)
    await self._send_command("READS CLEAR", allow_timeout=True)
    await self._send_command("POSITION CLEAR", allow_timeout=True)
    await self._send_command("MIRROR CLEAR", allow_timeout=True)

  async def _configure_absorbance(
    self,
    wavelength_nm: int,
    *,
    flashes: int,
    bandwidth: Optional[float] = None,
  ) -> None:
    wl_decitenth = int(round(wavelength_nm * 10))
    bw = bandwidth if bandwidth is not None else self._auto_bandwidth(wavelength_nm)
    bw_decitenth = int(round(bw * 10))
    reads_number = max(1, flashes)

    await self._send_command("MODE ABS")
    await self._clear_mode_settings(excitation=True)
    await self._send_command(
      f"EXCITATION 0,ABS,{wl_decitenth},{bw_decitenth},0", allow_timeout=True
    )
    await self._send_command(
      f"EXCITATION 1,ABS,{wl_decitenth},{bw_decitenth},0", allow_timeout=True
    )
    await self._send_command(f"READS 0,NUMBER={reads_number}", allow_timeout=True)
    await self._send_command(f"READS 1,NUMBER={reads_number}", allow_timeout=True)
    await self._send_command("TIME 0,READDELAY=0", allow_timeout=True)
    await self._send_command("TIME 1,READDELAY=0", allow_timeout=True)
    await self._send_command("SCAN DIRECTION=ALTUP", allow_timeout=True)
    await self._send_command("#RATIO LABELS", allow_timeout=True)
    await self._send_command(
      f"BEAM DIAMETER={self._capability_numeric('ABS', '#BEAM DIAMETER', 700)}", allow_timeout=True
    )
    await self._send_command("RATIO LABELS=1", allow_timeout=True)
    await self._send_command("PREPARE REF", allow_timeout=True, read_response=False)

  def _auto_bandwidth(self, wavelength_nm: int) -> float:
    """Return bandwidth in nm based on Infinite M specification."""

    return 9.0 if wavelength_nm > 315 else 5.0

  async def read_fluorescence(
    self,
    plate: Plate,
    wells: List[Well],
    excitation_wavelength: int,
    emission_wavelength: int,
    focal_height: float = 20.0,
    flashes: int = 25,
    integration_us: int = 20,
    gain: int = 100,
    excitation_bandwidth: int = 50,
    emission_bandwidth: int = 200,
    lag_us: int = 0,
  ) -> List[Dict]:
    """Queue and execute a fluorescence scan.

    Args:
      gain: PMT gain value (0-255).
      excitation_bandwidth: Excitation filter bandwidth in deci-tenths of nm.
      emission_bandwidth: Emission filter bandwidth in deci-tenths of nm.
      lag_us: Lag time in microseconds between excitation and measurement.
    """

    if not 230 <= excitation_wavelength <= 850:
      raise ValueError("Excitation wavelength must be between 230 nm and 850 nm.")
    if not 230 <= emission_wavelength <= 850:
      raise ValueError("Emission wavelength must be between 230 nm and 850 nm.")
    if focal_height < 0:
      raise ValueError("Focal height must be non-negative for fluorescence scans.")

    ordered_wells = wells if wells else plate.get_all_items()
    scan_wells = self._scan_visit_order(ordered_wells, serpentine=True)

    await self._begin_run()
    try:
      await self._configure_fluorescence(
        excitation_wavelength,
        emission_wavelength,
        focal_height,
        flashes=flashes,
        integration_us=integration_us,
        gain=gain,
        excitation_bandwidth=excitation_bandwidth,
        emission_bandwidth=emission_bandwidth,
        lag_us=lag_us,
      )
      decoder = _FluorescenceRunDecoder(len(scan_wells))

      await self._run_scan(
        ordered_wells=ordered_wells,
        decoder=decoder,
        mode="Fluorescence",
        step_loss_commands=[
          "CHECK MTP.STEPLOSS",
          "CHECK FI.TOP.STEPLOSS",
          "CHECK FI.STEPLOSS.Z",
        ],
        serpentine=True,
        scan_direction="UP",
      )

      if len(decoder.intensities) != len(scan_wells):
        raise RuntimeError("Fluorescence decoder did not complete scan.")
      intensities = decoder.intensities
      matrix = self._format_plate_result(plate, scan_wells, intensities)
      return [
        {
          "ex_wavelength": excitation_wavelength,
          "em_wavelength": emission_wavelength,
          "time": time.time(),
          "temperature": None,
          "data": matrix,
        }
      ]
    finally:
      await self._end_run()

  async def _configure_fluorescence(
    self,
    excitation_nm: int,
    emission_nm: int,
    focal_height: float,
    *,
    flashes: int,
    integration_us: int,
    gain: int,
    excitation_bandwidth: int,
    emission_bandwidth: int,
    lag_us: int,
  ) -> None:
    ex_decitenth = int(round(excitation_nm * 10))
    em_decitenth = int(round(emission_nm * 10))
    reads_number = max(1, flashes)
    beam_diameter = self._capability_numeric("FI.TOP", "#BEAM DIAMETER", 3000)
    z_position = int(round(focal_height * self.counts_per_mm_z))

    # UI issues the entire FI configuration twice before PREPARE REF.
    for _ in range(2):
      await self._send_command("MODE FI.TOP", allow_timeout=True)
      await self._clear_mode_settings(excitation=True, emission=True)
      await self._send_command(
        f"EXCITATION 0,FI,{ex_decitenth},{excitation_bandwidth},0", allow_timeout=True
      )
      await self._send_command(
        f"EMISSION 0,FI,{em_decitenth},{emission_bandwidth},0", allow_timeout=True
      )
      await self._send_command(f"TIME 0,INTEGRATION={integration_us}", allow_timeout=True)
      await self._send_command(f"TIME 0,LAG={lag_us}", allow_timeout=True)
      await self._send_command("TIME 0,READDELAY=0", allow_timeout=True)
      await self._send_command(f"GAIN 0,VALUE={gain}", allow_timeout=True)
      await self._send_command(f"POSITION 0,Z={z_position}", allow_timeout=True)
      await self._send_command(f"BEAM DIAMETER={beam_diameter}", allow_timeout=True)
      await self._send_command("SCAN DIRECTION=UP", allow_timeout=True)
      await self._send_command("RATIO LABELS=1", allow_timeout=True)
      await self._send_command(f"READS 0,NUMBER={reads_number}", allow_timeout=True)
      await self._send_command(
        f"EXCITATION 1,FI,{ex_decitenth},{excitation_bandwidth},0", allow_timeout=True
      )
      await self._send_command(
        f"EMISSION 1,FI,{em_decitenth},{emission_bandwidth},0", allow_timeout=True
      )
      await self._send_command(f"TIME 1,INTEGRATION={integration_us}", allow_timeout=True)
      await self._send_command(f"TIME 1,LAG={lag_us}", allow_timeout=True)
      await self._send_command("TIME 1,READDELAY=0", allow_timeout=True)
      await self._send_command(f"GAIN 1,VALUE={gain}", allow_timeout=True)
      await self._send_command(f"POSITION 1,Z={z_position}", allow_timeout=True)
      await self._send_command(f"READS 1,NUMBER={reads_number}", allow_timeout=True)
    await self._send_command("PREPARE REF", allow_timeout=True, read_response=False)

  async def read_luminescence(
    self,
    plate: Plate,
    wells: List[Well],
    focal_height: float = 20.0,
    flashes: int = 25,
    dark_integration_us: int = 3_000_000,
    meas_integration_us: int = 1_000_000,
  ) -> List[Dict]:
    """Queue and execute a luminescence scan."""

    if focal_height < 0:
      raise ValueError("Focal height must be non-negative for luminescence scans.")

    ordered_wells = wells if wells else plate.get_all_items()
    scan_wells = self._scan_visit_order(ordered_wells, serpentine=False)

    dark_integration = dark_integration_us
    meas_integration = meas_integration_us

    await self._begin_run()
    try:
      await self._configure_luminescence(
        dark_integration, meas_integration, focal_height, flashes=flashes
      )

      decoder = _LuminescenceRunDecoder(
        len(scan_wells),
        dark_integration_s=_integration_microseconds_to_seconds(dark_integration),
        meas_integration_s=_integration_microseconds_to_seconds(meas_integration),
      )

      await self._run_scan(
        ordered_wells=ordered_wells,
        decoder=decoder,
        mode="Luminescence",
        step_loss_commands=["CHECK MTP.STEPLOSS", "CHECK LUM.STEPLOSS"],
        serpentine=False,
        scan_direction="UP",
      )

      if len(decoder.measurements) != len(scan_wells):
        raise RuntimeError("Luminescence decoder did not complete scan.")
      intensities = [measurement.intensity for measurement in decoder.measurements]
      matrix = self._format_plate_result(plate, scan_wells, intensities)
      return [
        {
          "time": time.time(),
          "temperature": None,
          "data": matrix,
        }
      ]
    finally:
      await self._end_run()

  async def _await_measurements(
    self, decoder: "_MeasurementDecoder", row_count: int, mode: str
  ) -> None:
    target = decoder.count + row_count
    start_count = decoder.count
    self._drain_pending_bin_events(decoder)
    start = time.monotonic()
    reads = 0
    while decoder.count < target and (time.monotonic() - start) < self._max_row_wait_s:
      chunk = await self._read_packet(self._read_chunk_size)
      if not chunk:
        raise RuntimeError(f"{mode} read returned empty chunk; transport may not support reads.")
      decoder.feed(chunk)
      reads += 1
    if decoder.count < target:
      got = decoder.count - start_count
      raise RuntimeError(
        f"Timed out while parsing {mode.lower()} results "
        f"(decoded {got}/{row_count} measurements in {time.monotonic() - start:.1f}s, {reads} reads)."
      )

  def _drain_pending_bin_events(self, decoder: "_MeasurementDecoder") -> None:
    if not self._pending_bin_events:
      return
    for payload_len, blob in self._pending_bin_events:
      decoder.feed_bin(payload_len, blob)
    self._pending_bin_events.clear()

  async def _await_scan_terminal(self, saw_terminal: bool) -> None:
    if saw_terminal:
      return
    await self._read_command_response()

  async def _configure_luminescence(
    self,
    dark_integration: int,
    meas_integration: int,
    focal_height: float,
    *,
    flashes: int,
  ) -> None:
    await self._send_command("MODE LUM")
    # Pre-flight safety checks observed in captures (queries omitted).
    await self._send_command("CHECK LUM.FIBER")
    await self._send_command("CHECK LUM.LID")
    await self._send_command("CHECK LUM.STEPLOSS")
    await self._send_command("MODE LUM")
    reads_number = max(1, flashes)
    z_position = int(round(focal_height * self.counts_per_mm_z))
    await self._clear_mode_settings(emission=True)
    await self._send_command(f"POSITION LUM,Z={z_position}", allow_timeout=True)
    await self._send_command(f"TIME 0,INTEGRATION={dark_integration}", allow_timeout=True)
    await self._send_command(f"READS 0,NUMBER={reads_number}", allow_timeout=True)
    await self._send_command("SCAN DIRECTION=UP", allow_timeout=True)
    await self._send_command("RATIO LABELS=1", allow_timeout=True)
    await self._send_command("EMISSION 1,EMPTY,0,0,0", allow_timeout=True)
    await self._send_command(f"TIME 1,INTEGRATION={meas_integration}", allow_timeout=True)
    await self._send_command("TIME 1,READDELAY=0", allow_timeout=True)
    await self._send_command(f"READS 1,NUMBER={reads_number}", allow_timeout=True)
    await self._send_command("#EMISSION ATTENUATION", allow_timeout=True)
    await self._send_command("PREPARE REF", allow_timeout=True, read_response=False)

  def _group_by_row(self, wells: Sequence[Well]) -> List[Tuple[int, List[Well]]]:
    grouped: Dict[int, List[Well]] = {}
    for well in wells:
      grouped.setdefault(well.get_row(), []).append(well)
    for row in grouped.values():
      row.sort(key=lambda w: w.get_column())
    return sorted(grouped.items(), key=lambda item: item[0])

  def _scan_visit_order(self, wells: Sequence[Well], serpentine: bool) -> List[Well]:
    visit: List[Well] = []
    for row_index, row_wells in self._group_by_row(wells):
      if serpentine and row_index % 2 == 1:
        visit.extend(reversed(row_wells))
      else:
        visit.extend(row_wells)
    return visit

  def _map_well_to_stage(self, well: Well) -> StagePosition:
    if well.location is None:
      raise ValueError("Well does not have a location assigned within its plate definition.")
    center = well.location + well.get_anchor(x="c", y="c")
    stage_x = int(round(center.x * self.counts_per_mm_x))
    parent_plate = well.parent
    if parent_plate is None or not isinstance(parent_plate, Plate):
      raise ValueError("Well is not assigned to a plate; cannot derive stage coordinates.")
    plate_height_mm = parent_plate.get_size_y()
    stage_y = int(round((plate_height_mm - center.y) * self.counts_per_mm_y))
    return stage_x, stage_y

  def _scan_range(
    self, row_index: int, row_wells: Sequence[Well], serpentine: bool
  ) -> Tuple[int, int, int]:
    """Return start/end/count for a row, honoring serpentine layout when requested."""

    first_x, _ = self._map_well_to_stage(row_wells[0])
    last_x, _ = self._map_well_to_stage(row_wells[-1])
    count = len(row_wells)
    if not serpentine:
      return min(first_x, last_x), max(first_x, last_x), count
    if row_index % 2 == 0:
      return first_x, last_x, count
    return last_x, first_x, count

  def _format_plate_result(
    self, plate: Plate, wells: Sequence[Well], values: Sequence[float]
  ) -> List[List[Optional[float]]]:
    matrix: List[List[Optional[float]]] = [
      [None for _ in range(plate.num_items_x)] for _ in range(plate.num_items_y)
    ]
    for well, val in zip(wells, values):
      r, c = well.get_row(), well.get_column()
      if 0 <= r < plate.num_items_y and 0 <= c < plate.num_items_x:
        matrix[r][c] = float(val)
    return matrix

  async def _initialize_device(self) -> None:
    try:
      await self._send_command("QQ")
    except TimeoutError:
      logger.warning("QQ produced no response; continuing with initialization.")
    await self._send_command("INIT FORCE")

  async def _begin_run(self) -> None:
    self._reset_stream_state()
    await self._send_command("KEYLOCK ON")
    self._run_active = True

  def _reset_stream_state(self) -> None:
    self._pending_bin_events.clear()
    self._parser = _StreamParser(allow_bare_ascii=True)

  async def _read_packet(self, size: int) -> bytes:
    try:
      data = await self.io.read(size=size)
    except TimeoutError:
      await self._recover_transport()
      raise
    return data

  async def _recover_transport(self) -> None:
    try:
      await self.io.stop()
      await asyncio.sleep(0.2)
      await self.io.setup()
    except Exception:
      logger.warning("Transport recovery failed.", exc_info=True)
      return
    self._mode_capabilities.clear()
    self._reset_stream_state()
    await self._initialize_device()

  async def _end_run(self) -> None:
    try:
      await self._send_command("TERMINATE", allow_timeout=True)
      for cmd in self._active_step_loss_commands:
        await self._send_command(cmd, allow_timeout=True)
      await self._send_command("KEYLOCK OFF", allow_timeout=True)
      await self._send_command("ABSOLUTE MTP,IN", allow_timeout=True)
    finally:
      self._run_active = False
      self._active_step_loss_commands = []

  async def _cleanup_protocol(self) -> None:
    async def send_cleanup_cmd(cmd: str) -> None:
      try:
        await self._send_command(cmd, allow_timeout=True, read_response=False)
      except Exception:
        logger.warning("Cleanup command failed: %s", cmd)

    if self._run_active or self._active_step_loss_commands:
      await send_cleanup_cmd("TERMINATE")
      for cmd in self._active_step_loss_commands:
        await send_cleanup_cmd(cmd)
    await send_cleanup_cmd("KEYLOCK OFF")
    await send_cleanup_cmd("ABSOLUTE MTP,IN")
    self._run_active = False
    self._active_step_loss_commands = []

  async def _query_mode_capabilities(self, mode: str) -> None:
    commands = self._MODE_CAPABILITY_COMMANDS.get(mode)
    if not commands:
      return
    try:
      await self._send_command(f"MODE {mode}")
    except TimeoutError:
      logger.warning("Capability MODE %s timed out; continuing without mode capabilities.", mode)
      return
    collected: Dict[str, str] = {}
    for cmd in commands:
      try:
        frames = await self._send_command(cmd)
      except TimeoutError:
        logger.warning("Capability query '%s' timed out; proceeding with defaults.", cmd)
        continue
      if frames:
        collected[cmd] = frames[-1]
    if collected:
      self._mode_capabilities[mode] = collected

  def _get_mode_capability(self, mode: str, command: str) -> Optional[str]:
    return self._mode_capabilities.get(mode, {}).get(command)

  def _capability_numeric(self, mode: str, command: str, fallback: int) -> int:
    resp = self._get_mode_capability(mode, command)
    if not resp:
      return fallback
    token = resp.split("|")[0].split(":")[0].split("~")[0].strip()
    if not token:
      return fallback
    try:
      return int(float(token))
    except ValueError:
      return fallback

  @staticmethod
  def _frame_command(command: str) -> bytes:
    """Return a framed command with length/checksum trailer."""

    payload = command.encode("ascii")
    xor = 0
    for byte in payload:
      xor ^= byte
    checksum = (xor ^ 0x01) & 0xFF
    length = len(payload) & 0xFF
    return b"\x02" + payload + b"\x03\x00\x00" + bytes([length, checksum]) + b"\x0d"

  async def _send_command(
    self,
    command: str,
    wait_for_terminal: bool = True,
    allow_timeout: bool = False,
    read_response: bool = True,
  ) -> List[str]:
    logger.debug("[tecan] >> %s", command)
    framed = self._frame_command(command)
    await self.io.write(framed)
    if not read_response:
      return []
    if command.startswith(("#", "?")):
      try:
        return await self._read_command_response(require_terminal=False)
      except TimeoutError:
        if allow_timeout:
          logger.warning("Timeout waiting for response to %s", command)
          return []
        raise
    try:
      frames = await self._read_command_response(require_terminal=wait_for_terminal)
    except TimeoutError:
      if allow_timeout:
        logger.warning("Timeout waiting for response to %s", command)
        return []
      raise
    for pkt in frames:
      logger.debug("[tecan] << %s", pkt)
    return frames

  async def _drain(self, attempts: int = 4) -> None:
    """Read and discard a few packets to clear the stream."""
    for _ in range(attempts):
      data = await self._read_packet(128)
      if not data:
        break

  async def _read_command_response(
    self, max_iterations: int = 8, require_terminal: bool = True
  ) -> List[str]:
    """Read response frames and cache any binary payloads that arrive."""
    frames: List[str] = []
    saw_terminal = False
    for _ in range(max_iterations):
      chunk = await self._read_packet(128)
      if not chunk:
        break
      for event in self._parser.feed(chunk):
        if event.text is not None:
          frames.append(event.text)
          if self._is_terminal_frame(event.text):
            saw_terminal = True
        elif event.payload_len is not None and event.blob is not None:
          self._pending_bin_events.append((event.payload_len, event.blob))
      if not require_terminal and frames and not self._parser.has_pending_bin():
        break
      if require_terminal and saw_terminal and not self._parser.has_pending_bin():
        break
    if require_terminal and not saw_terminal:
      # best effort: drain once more so pending ST doesn't leak into next command
      await self._drain(1)
    return frames

  @staticmethod
  def _is_terminal_frame(text: str) -> bool:
    """Return True if the ASCII frame is a terminal marker."""
    return text in {"ST", "+", "-"} or text.startswith("BY#T")


@dataclass
class _AbsorbanceMeasurement:
  sample: int
  reference: int
  items: Optional[List[Tuple[int, int]]] = None


class _AbsorbanceRunDecoder(_MeasurementDecoder):
  """Incrementally decode absorbance measurement frames."""

  STATUS_FRAME_LEN = 31

  def __init__(self, expected: int) -> None:
    super().__init__(expected)
    self.measurements: List[_AbsorbanceMeasurement] = []
    self._calibration: Optional[_AbsorbanceCalibration] = None

  @property
  def count(self) -> int:
    return len(self.measurements)

  @property
  def calibration(self) -> Optional[_AbsorbanceCalibration]:
    """Return the absorbance calibration data, if available."""
    return self._calibration

  def _should_consume_bin(self, payload_len: int) -> bool:
    return _is_abs_calibration_len(payload_len) or _is_abs_data_len(payload_len)

  def _handle_bin(self, payload_len: int, blob: bytes) -> None:
    if _is_abs_calibration_len(payload_len):
      if self._calibration is not None:
        return
      cal = _decode_abs_calibration(payload_len, blob)
      if cal is not None:
        self._calibration = cal
      return
    if _is_abs_data_len(payload_len):
      data = _decode_abs_data(payload_len, blob)
      if data is None:
        return
      _label, _ex, items = data
      sample, reference = items[0] if items else (0, 0)
      self.measurements.append(
        _AbsorbanceMeasurement(sample=sample, reference=reference, items=items)
      )


class _FluorescenceRunDecoder(_MeasurementDecoder):
  """Incrementally decode fluorescence measurement frames."""

  STATUS_FRAME_LEN = 31

  def __init__(self, expected_wells: int) -> None:
    super().__init__(expected_wells)
    self._intensities: List[int] = []
    self._calibration: Optional[_FluorescenceCalibration] = None

  @property
  def count(self) -> int:
    return len(self._intensities)

  @property
  def intensities(self) -> List[int]:
    """Return decoded fluorescence intensities."""
    return self._intensities

  def _should_consume_bin(self, payload_len: int) -> bool:
    if payload_len == 18:
      return True
    if payload_len >= 16 and (payload_len - 6) % 10 == 0:
      return True
    return False

  def _handle_bin(self, payload_len: int, blob: bytes) -> None:
    if payload_len == 18:
      cal = _decode_flr_calibration(payload_len, blob)
      if cal is not None:
        self._calibration = cal
      return
    data = _decode_flr_data(payload_len, blob)
    if data is None:
      return
    _label, _ex, _em, items = data
    if self._calibration is not None:
      intensity = _fluorescence_corrected(self._calibration, items)
    else:
      if not items:
        intensity = 0
      else:
        intensity = int(round(sum(m for m, _ in items) / len(items)))
    self._intensities.append(intensity)


@dataclass
class _LuminescenceMeasurement:
  intensity: int


class _LuminescenceRunDecoder(_MeasurementDecoder):
  """Incrementally decode luminescence measurement frames."""

  def __init__(
    self,
    expected: int,
    *,
    dark_integration_s: float = 0.0,
    meas_integration_s: float = 0.0,
  ) -> None:
    super().__init__(expected)
    self.measurements: List[_LuminescenceMeasurement] = []
    self._calibration: Optional[_LuminescenceCalibration] = None
    self._dark_integration_s = float(dark_integration_s)
    self._meas_integration_s = float(meas_integration_s)

  @property
  def count(self) -> int:
    return len(self.measurements)

  def _should_consume_bin(self, payload_len: int) -> bool:
    if payload_len == 10:
      return True
    if payload_len >= 14 and (payload_len - 4) % 10 == 0:
      return True
    return False

  def _handle_bin(self, payload_len: int, blob: bytes) -> None:
    if payload_len == 10:
      cal = _decode_lum_calibration(payload_len, blob)
      if cal is not None:
        self._calibration = cal
      return
    data = _decode_lum_data(payload_len, blob)
    if data is None:
      return
    _label, _em, counts = data
    if self._calibration is not None and self._dark_integration_s and self._meas_integration_s:
      intensity = _luminescence_intensity(
        self._calibration, counts, self._dark_integration_s, self._meas_integration_s
      )
    else:
      intensity = int(round(sum(counts) / len(counts))) if counts else 0
    self.measurements.append(_LuminescenceMeasurement(intensity=intensity))


__all__ = [
  "ExperimentalTecanInfinite200ProBackend",
]



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Optional, Union, cast

from pylabrobot.liquid_handling.backends.backend import (
  LiquidHandlerBackend,
)
from pylabrobot.liquid_handling.standard import (
  Drop,
  DropTipRack,
  MultiHeadAspirationContainer,
  MultiHeadAspirationPlate,
  MultiHeadDispenseContainer,
  MultiHeadDispensePlate,
  Pickup,
  PickupTipRack,
  ResourceDrop,
  ResourceMove,
  ResourcePickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)
from pylabrobot.resources import Tip
from pylabrobot.serializer import serialize


class SerializingBackend(LiquidHandlerBackend, metaclass=ABCMeta):
  """A backend that serializes all commands received, and sends them to `self.send_command` for
  processing. The implementation of `send_command` is left to the subclasses."""

  def __init__(self, num_channels: int):
    LiquidHandlerBackend.__init__(self)
    self._num_channels = num_channels
    self._num_arms = 1
    self._head96_installed = True

  @property
  def num_channels(self) -> int:
    return self._num_channels

  @abstractmethod
  async def send_command(
    self, command: str, data: Optional[Dict[str, Any]] = None
  ) -> Optional[dict]:
    raise NotImplementedError

  async def setup(self):
    await super().setup()
    await self.send_command(command="setup")

  async def stop(self):
    await self.send_command(command="stop")

  def serialize(self) -> dict:
    return {**super().serialize(), "num_channels": self.num_channels}

  async def pick_up_tips(self, ops: List[Pickup], use_channels: List[int]):
    serialized = [
      {
        "resource_name": op.resource.name,
        "offset": serialize(op.offset),
        "tip": op.tip.serialize(),
      }
      for op in ops
    ]
    await self.send_command(
      command="pick_up_tips",
      data={"channels": serialized, "use_channels": use_channels},
    )

  async def drop_tips(self, ops: List[Drop], use_channels: List[int]):
    serialized = [
      {
        "resource_name": op.resource.name,
        "offset": serialize(op.offset),
        "tip": op.tip.serialize(),
      }
      for op in ops
    ]
    await self.send_command(
      command="drop_tips",
      data={"channels": serialized, "use_channels": use_channels},
    )

  async def aspirate(self, ops: List[SingleChannelAspiration], use_channels: List[int]):
    serialized = [
      {
        "resource_name": op.resource.name,
        "offset": serialize(op.offset),
        "tip": serialize(op.tip),
        "volume": op.volume,
        "flow_rate": serialize(op.flow_rate),
        "liquid_height": serialize(op.liquid_height),
        "blow_out_air_volume": serialize(op.blow_out_air_volume),
        "mix": serialize(op.mix),
      }
      for op in ops
    ]
    await self.send_command(
      command="aspirate",
      data={"channels": serialized, "use_channels": use_channels},
    )

  async def dispense(self, ops: List[SingleChannelDispense], use_channels: List[int]):
    serialized = [
      {
        "resource_name": op.resource.name,
        "offset": serialize(op.offset),
        "tip": serialize(op.tip),
        "volume": op.volume,
        "flow_rate": serialize(op.flow_rate),
        "liquid_height": serialize(op.liquid_height),
        "blow_out_air_volume": serialize(op.blow_out_air_volume),
        "mix": serialize(op.mix),
      }
      for op in ops
    ]
    await self.send_command(
      command="dispense",
      data={"channels": serialized, "use_channels": use_channels},
    )

  async def pick_up_tips96(self, pickup: PickupTipRack):
    await self.send_command(
      command="pick_up_tips96",
      data={
        "resource_name": pickup.resource.name,
        "offset": serialize(pickup.offset),
      },
    )

  async def drop_tips96(self, drop: DropTipRack):
    await self.send_command(
      command="drop_tips96",
      data={
        "resource_name": drop.resource.name,
        "offset": serialize(drop.offset),
      },
    )

  async def aspirate96(
    self, aspiration: Union[MultiHeadAspirationPlate, MultiHeadAspirationContainer]
  ):
    data = {
      "aspiration": {
        "offset": serialize(aspiration.offset),
        "volume": aspiration.volume,
        "flow_rate": serialize(aspiration.flow_rate),
        "liquid_height": serialize(aspiration.liquid_height),
        "blow_out_air_volume": serialize(aspiration.blow_out_air_volume),
        "tips": [serialize(tip) for tip in aspiration.tips],
      }
    }
    if isinstance(aspiration, MultiHeadAspirationPlate):
      data["aspiration"]["well_names"] = [well.name for well in aspiration.wells]
    else:
      data["aspiration"]["trough"] = aspiration.container.name
    await self.send_command(command="aspirate96", data=data)

  async def dispense96(self, dispense: Union[MultiHeadDispensePlate, MultiHeadDispenseContainer]):
    data = {
      "dispense": {
        "offset": serialize(dispense.offset),
        "volume": dispense.volume,
        "flow_rate": serialize(dispense.flow_rate),
        "liquid_height": serialize(dispense.liquid_height),
        "blow_out_air_volume": serialize(dispense.blow_out_air_volume),
        "tips": [serialize(tip) for tip in dispense.tips],
      }
    }
    if isinstance(dispense, MultiHeadDispensePlate):
      data["dispense"]["well_names"] = [well.name for well in dispense.wells]
    else:
      data["dispense"]["trough"] = dispense.container.name
    await self.send_command(command="dispense96", data=data)

  async def pick_up_resource(self, pickup: ResourcePickup, **backend_kwargs):
    await self.send_command(
      command="pick_up_resource",
      data={
        "resource_name": pickup.resource.name,
        "offset": serialize(pickup.offset),
        "pickup_distance_from_top": pickup.pickup_distance_from_top,
        "direction": serialize(pickup.direction),
      },
      **backend_kwargs,
    )

  async def move_picked_up_resource(self, move: ResourceMove, **backend_kwargs):
    await self.send_command(
      command="move_picked_up_resource",
      data={
        "resource_name": move.resource.name,
        "location": serialize(move.location),
        "gripped_direction": serialize(move.gripped_direction),
      },
      **backend_kwargs,
    )

  async def drop_resource(self, drop: ResourceDrop, **backend_kwargs):
    await self.send_command(
      command="drop_resource",
      data={
        "resource_name": drop.resource.name,
        "destination": serialize(drop.destination),
        "offset": serialize(drop.offset),
        "pickup_distance_from_top": drop.pickup_distance_from_top,
        "pickup_direction": serialize(drop.pickup_direction),
        "drop_direction": serialize(drop.direction),
        "rotation": drop.rotation,
      },
      **backend_kwargs,
    )

  async def prepare_for_manual_channel_operation(self, channel: int):
    await self.send_command(
      command="prepare_for_manual_channel_operation",
      data={"channel": channel},
    )

  async def move_channel_x(self, channel: int, x: float):
    await self.send_command(command="move_channel_x", data={"channel": channel, "x": x})

  async def move_channel_y(self, channel: int, y: float):
    await self.send_command(command="move_channel_y", data={"channel": channel, "y": y})

  async def move_channel_z(self, channel: int, z: float):
    await self.send_command(command="move_channel_z", data={"channel": channel, "z": z})

  async def request_tip_presence(self) -> List[Optional[bool]]:
    """Request tip presence on each channel via the serialized command interface.

    Returns:
      A list of length `num_channels` where each element is `True` if a tip is mounted,
      `False` if not, or `None` if unknown.
    """
    result = await self.send_command(command="request_tip_presence")
    if result is not None and "tip_presence" in result:
      return cast(List[Optional[bool]], result["tip_presence"])
    return [None] * self.num_channels

  def can_pick_up_tip(self, channel_idx: int, tip: Tip) -> bool:
    return True



import unittest

from pylabrobot.resources import PlateAdapter

_adapter = PlateAdapter(
  name="adapter",
  size_x=128.0,
  size_y=86.0,
  size_z=15.0,
  dx=0.0,
  dy=1.0,
  dz=2.0,
  adapter_hole_size_x=100.0,
  adapter_hole_size_y=80.0,
  adapter_hole_dx=9.0,
  adapter_hole_dy=9.0,
  plate_z_offset=3.0,
)


class TestPlateAdapter(unittest.TestCase):
  def test_plate_adapter_serialization(self):
    serialized = _adapter.serialize()
    self.assertEqual(
      serialized,
      {
        "name": "adapter",
        "type": "PlateAdapter",
        "size_x": 128.0,
        "size_y": 86.0,
        "size_z": 15.0,
        "location": None,
        "rotation": {"x": 0, "y": 0, "z": 0, "type": "Rotation"},
        "category": "plate_adapter",
        "model": None,
        "barcode": None,
        "preferred_pickup_location": None,
        "children": [],
        "parent_name": None,
        "dx": 0.0,
        "dy": 1.0,
        "dz": 2.0,
        "adapter_hole_size_x": 100.0,
        "adapter_hole_size_y": 80.0,
        "adapter_hole_dx": 9.0,
        "adapter_hole_dy": 9.0,
        "plate_z_offset": 3.0,
      },
    )

  def test_plate_adapter_deserialization(self):
    assert _adapter == PlateAdapter.deserialize(_adapter.serialize())



import unittest
import unittest.mock
from unittest.mock import call

from pylabrobot.liquid_handling.backends.tecan.EVO_backend import (
  EVOBackend,
  LiHa,
  RoMa,
)
from pylabrobot.liquid_handling.standard import (
  GripDirection,
  Pickup,
  ResourceDrop,
  ResourcePickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)
from pylabrobot.resources import (
  Coordinate,
  DeepWell_96_Well,
  DiTi_100ul_Te_MO,
  DiTi_SBS_3_Pos_MCA96,
  EVO150Deck,
  MP_3Pos_PCR,
)
from pylabrobot.resources.rotation import Rotation


class EVOTests(unittest.IsolatedAsyncioTestCase):
  """Test that the EVO backend is calling `send_command` correctly."""

  def setUp(self) -> None:
    super().setUp()

    # mock the EVO
    self.evo = EVOBackend(diti_count=8)
    self.evo.send_command = unittest.mock.AsyncMock()  # type: ignore[method-assign]

    async def send_command(module, command, params=None):
      if command == "RPX":  # report_x_param
        return {"data": [9000]}  # park position roma
      if command == "RPY":  # report_y_param
        return {"data": [90]}  # park position roma
      if command == "RPZ":  # report_z_param
        return {"data": [2000]}  # dummy value
      return {"data": None}

    self.evo.send_command.side_effect = send_command  # type: ignore[method-assign]

    self.deck = EVO150Deck()
    self.evo.set_deck(self.deck)

    # setup
    self.evo.setup = unittest.mock.AsyncMock()  # type: ignore[method-assign]
    self.evo._num_channels = 8
    self.evo._x_range = 2000  # TODO: override report_x_param
    self.evo._y_range = 2000
    self.evo._z_range = 2000
    self.evo._roma_connected = True
    self.evo._liha_connected = True
    self.evo.liha = LiHa(self.evo, EVOBackend.LIHA)
    self.evo.roma = RoMa(self.evo, EVOBackend.ROMA)

    # deck setup
    self.tr_carrier = DiTi_SBS_3_Pos_MCA96(name="tip_rack_carrier")
    self.tr_carrier[0] = self.tr = DiTi_100ul_Te_MO(name="tip_rack")
    self.deck.assign_child_resource(self.tr_carrier, rails=10)

    self.plate_carrier = MP_3Pos_PCR(name="plate_carrier")
    self.plate_carrier[0] = self.plate = DeepWell_96_Well(name="plate")
    self.deck.assign_child_resource(self.plate_carrier, rails=16)

    self.evo.send_command.reset_mock()

  async def test_pick_up_tip(self):
    op = Pickup(
      resource=self.tr.get_item("A1"),
      offset=Coordinate.zero(),
      tip=self.tr.get_tip("A1"),
    )
    await self.evo.pick_up_tips([op], use_channels=[0])

    self.evo.send_command.assert_has_calls(  # type: ignore[attr-defined]
      [
        call(
          module="C5",
          command="SHZ",
          params=[2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000],
        ),
        call(module="C5", command="RPX", params=[0]),
        call(
          module="C5",
          command="PAA",
          params=[
            2380,
            1991,
            90,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
          ],
        ),
        call(
          module="C5",
          command="PVL",
          params=[0, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SEP",
          params=[420, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="PPR",
          params=[30, None, None, None, None, None, None, None],
        ),
        call(module="C5", command="AGT", params=[1, 1023, 210, 0]),
      ]
    )

  # TODO: add Trash to Tecan deck or allow dropping tips in a non-Trash area
  # async def test_drop_tip(self):
  #   op = Drop(
  #     resource=self.deck.get_trash_area(),
  #     offset=None,
  #     tip=self.tr.get_tip("A1")
  #   )
  #   await self.evo.drop_tips([op], use_channels=[0])
  #   self.evo.send_command.assert_has_calls([])

  async def test_aspirate(self):
    op = SingleChannelAspiration(
      resource=self.plate.get_item("A1"),
      offset=Coordinate.zero(),
      tip=self.tr.get_tip("A1"),
      volume=100,
      flow_rate=100,
      liquid_height=10,
      blow_out_air_volume=0,
      mix=None,
    )
    await self.evo.aspirate([op], use_channels=[0])
    self.evo.send_command.assert_has_calls(  # type: ignore[attr-defined]
      [
        call(
          module="C5",
          command="SHZ",
          params=[2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000],
        ),
        call(module="C5", command="RPX", params=[0]),
        call(
          module="C5",
          command="PAA",
          params=[
            3829,
            2051,
            90,
            2100,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
          ],
        ),
        call(
          module="C5",
          command="PVL",
          params=[0, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SEP",
          params=[420, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="PPR",
          params=[15, None, None, None, None, None, None, None],
        ),
        call(module="C5", command="SDM", params=[7, 1]),
        call(
          module="C5",
          command="SSL",
          params=[600, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SDL",
          params=[40, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="STL",
          params=[1375, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SML",
          params=[985, 2000, 2000, 2000, 2000, 2000, 2000, 2000],
        ),
        call(
          module="C5",
          command="SBL",
          params=[20, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SHZ",
          params=[2100, 2100, 2100, 2100, 2100, 2100, 2100, 2100],
        ),
        call(
          module="C5",
          command="MDT",
          params=[1, None, None, None, 30, 0, 0, 0, 0, 0, 0, 0],
        ),
        call(
          module="C5",
          command="SHZ",
          params=[2000, 2000, 2000, 2000, 2000, 2000, 2000, 2000],
        ),
        call(
          module="C5",
          command="SSZ",
          params=[30, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SEP",
          params=[600, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="STZ",
          params=[-30, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="MTR",
          params=[313, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SSZ",
          params=[200, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="MAZ",
          params=[1375, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="PVL",
          params=[0, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SEP",
          params=[420, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="PPR",
          params=[30, None, None, None, None, None, None, None],
        ),
      ]
    )

  async def test_dispense(self):
    op = SingleChannelDispense(
      resource=self.plate.get_item("A1"),
      offset=Coordinate.zero(),
      tip=self.tr.get_tip("A1"),
      volume=100,
      flow_rate=100,
      liquid_height=10,
      blow_out_air_volume=0,
      mix=None,
    )
    await self.evo.dispense([op], use_channels=[0])
    self.evo.send_command.assert_has_calls(  # type: ignore[attr-defined]
      [
        call(module="C5", command="RPX", params=[0]),
        call(
          module="C5",
          command="PAA",
          params=[
            3829,
            2051,
            90,
            1355,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
            2000,
          ],
        ),
        call(
          module="C5",
          command="SEP",
          params=[600, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="SPP",
          params=[2400, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="STZ",
          params=[0, None, None, None, None, None, None, None],
        ),
        call(
          module="C5",
          command="MTR",
          params=[-358, None, None, None, None, None, None, None],
        ),
      ]
    )

  # async def test_aspirate_custom_flow_rate(self):
  #   op = SingleChannelAspiration(
  #     resource=self.plate.get_item("A1"),
  #     offset=Coordinate.zero(),
  #     tip=self.tr.get_tip("A1"),
  #     volume=100,
  #     flow_rate=200,
  #     liquid_height=10,
  #     blow_out_air_volume=0,
  #     liquids=[(None, 100)],
  #   )
  #   await self.evo.aspirate([op], use_channels=[0])
  #   self.evo.send_command.assert_any_call(  # type: ignore[attr-defined]
  #     module="C5",
  #     command="SSZ",
  #     params=[60, None, None, None, None, None, None, None],
  #   )
  #   self.evo.send_command.assert_any_call(  # type: ignore[attr-defined]
  #     module="C5",
  #     command="SEP",
  #     params=[2400, None, None, None, None, None, None, None],
  #   )

  # async def test_dispense_custom_flow_rate(self):
  #   op = SingleChannelDispense(
  #     resource=self.plate.get_item("A1"),
  #     offset=Coordinate.zero(),
  #     tip=self.tr.get_tip("A1"),
  #     volume=100,
  #     flow_rate=200,
  #     liquid_height=10,
  #     blow_out_air_volume=0,
  #     liquids=[(None, 100)],
  #   )
  #   await self.evo.dispense([op], use_channels=[0])
  #   self.evo.send_command.assert_any_call(  # type: ignore[attr-defined]
  #     module="C5",
  #     command="SEP",
  #     params=[2400, None, None, None, None, None, None, None],
  #   )

  async def test_move_resource(self):
    pickup = ResourcePickup(
      resource=self.plate,
      offset=Coordinate.zero(),
      pickup_distance_from_top=13.2,
      direction=GripDirection.FRONT,
    )
    drop = ResourceDrop(
      resource=self.plate,
      destination=self.plate_carrier[0].get_location_wrt(self.deck),
      destination_absolute_rotation=Rotation(0, 0, 0),
      offset=Coordinate.zero(),
      pickup_distance_from_top=13.2,
      pickup_direction=GripDirection.FRONT,
      direction=GripDirection.FRONT,
      rotation=0,
    )

    await self.evo.pick_up_resource(pickup)
    await self.evo.drop_resource(drop)

    self.evo.send_command.assert_has_calls(  # type: ignore[attr-defined]
      [
        call(module="C1", command="RPZ", params=[5]),
        call(module="C1", command="SSM", params=[1]),
        call(module="C1", command="SFX", params=[10000, None]),
        call(module="C1", command="SFY", params=[5000, 1500]),
        call(module="C1", command="SFZ", params=[1300, None]),
        call(module="C1", command="SFR", params=[5000, 1500]),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[1, 5561, 2368, 1054, 900, None, 1, 0, 0],
        ),
        call(module="C1", command="AAC"),
        call(module="C1", command="RPX", params=[0]),
        call(module="C1", command="SSM", params=[0]),
        call(module="C1", command="PAG", params=[900]),
        call(module="C1", command="STW", params=[1, 0, 0, 0, 135, 0]),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[1, 5561, 2368, 687, 900, None, 1, 0, 1],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[1, 5561, 2368, 59, 900, None, 1, 0, 0],
        ),
        call(module="C1", command="AAC"),
        call(module="C1", command="RPX", params=[0]),
        call(module="C1", command="SFY", params=[3500, 1000]),
        call(module="C1", command="SFR", params=[2000, 600]),
        call(module="C1", command="SGG", params=[100, 75, None]),
        call(module="C1", command="AGR", params=[754]),
        call(module="C1", command="RPZ", params=[5]),
        call(module="C1", command="STW", params=[1, 0, 0, 0, 135, 0]),
        call(module="C1", command="STW", params=[2, 0, 0, 0, 53, 0]),
        call(module="C1", command="STW", params=[3, 0, 0, 0, 55, 0]),
        call(module="C1", command="STW", params=[4, 45, 0, 0, 0, 0]),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[1, 5561, 2368, 59, 900, None, 1, 0, 1],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[2, 5561, 2368, 687, 900, None, 1, 0, 2],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[3, 5561, 2368, 1054, 900, None, 1, 0, 3],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[4, 5561, 2368, 1054, 900, None, 1, 0, 4],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[5, 5561, 2368, 687, 900, None, 1, 0, 3],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[6, 5561, 2368, 59, 900, None, 1, 0, 0],
        ),
        call(module="C1", command="AAC"),
        call(module="C1", command="RPX", params=[0]),
        call(module="C1", command="PAG", params=[900]),
        call(module="C1", command="SFY", params=[5000, 1500]),
        call(module="C1", command="SFR", params=[5000, 1500]),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[1, 5561, 2368, 59, 900, None, 1, 0, 1],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[2, 5561, 2368, 687, 900, None, 1, 0, 2],
        ),
        call(module="C1", command="RPX", params=[0]),
        call(
          module="C1",
          command="SAA",
          params=[3, 5561, 2368, 1054, 900, None, 1, 0, 0],
        ),
        call(module="C1", command="AAC"),
        call(module="C1", command="RPX", params=[0]),
        call(module="C1", command="SFY", params=[3500, 1000]),
        call(module="C1", command="SFR", params=[2000, 600]),
      ]
    )



import asyncio
import dataclasses
import enum
import time
from typing import Set

try:
  import serial

  HAS_SERIAL = True
except ImportError as e:
  HAS_SERIAL = False
  _SERIAL_IMPORT_ERROR = e

from pylabrobot.io.serial import Serial
from pylabrobot.sealing.backend import SealerBackend


class A4SBackend(SealerBackend):
  def __init__(self, port: str, timeout=20) -> None:
    if not HAS_SERIAL:
      raise RuntimeError(
        "pyserial is not installed. Install with: pip install pylabrobot[serial]. "
        f"Import error: {_SERIAL_IMPORT_ERROR}"
      )
    super().__init__()
    self.port = port
    self.timeout = timeout
    self.io = Serial(
      port=self.port,
      baudrate=19200,
      bytesize=serial.EIGHTBITS,
      parity=serial.PARITY_NONE,
      stopbits=serial.STOPBITS_ONE,
      human_readable_device_name="A4S Sealer",
    )

  async def setup(self):
    await self.io.setup()
    await self.system_reset()

  async def stop(self):
    await self.set_heater(on=False)
    await self.io.stop()

  async def set_heater(self, on: bool):
    """Set the heater on or off."""
    command = "*00H1ZZ" if on else "*00H0ZZ"
    await self.send_command(command)
    return await self._wait_for_status({A4SBackend.Status.SystemStatus.idle})

  @dataclasses.dataclass
  class Status:
    class SystemStatus(enum.Enum):
      idle = 0
      single_cycle = 1
      repeat_cycle = 2
      error = 3
      finish = 4

    class HeaterBlockStatus(enum.Enum):
      heater_off = 0
      ready = 1
      heating = 2
      cooling = 3
      converging = 4

    @dataclasses.dataclass
    class SensorStatus:
      shuttle_middle_sensor: bool
      shuttle_open_sensor: bool
      shuttle_close_sensor: bool
      clean_door_sensor: bool
      seal_roll_sensor: bool
      heater_motor_up_sensor: bool
      heater_motor_down_sensor: bool
      # no_connect: bool

    current_temperature: float
    system_status: SystemStatus
    heater_block_status: HeaterBlockStatus
    error_code: int
    warning_code: int
    sensor_status: SensorStatus
    remaining_time: int

  async def _read_message(self) -> str:
    """read a message. we are not sure what format it is."""
    start = time.time()
    r, x = b"", b""
    has_read_r = False
    while x != b"" or (len(r) == 0 and x == b""):
      x = await self.io.read()
      if has_read_r:
        r += x
      if x == b"\r":
        if not has_read_r:
          has_read_r = True
        else:
          break
      if time.time() - start > self.timeout:
        raise TimeoutError("Timeout while waiting for response")
    return r.decode("utf-8")

  async def get_status(self) -> Status:
    # read until we get a system status message
    message: str
    while True:
      message = await self._read_message()
      if message[1] == "T":  # read system status
        break
      # message[1] == b"D": # Operation Status Message Format
      # message[1] == b"Y": # Response of Command Accepted Message Format
      # message[1] == b"N": # Response of Command Rejected Message Format
      # message[1] == b"X": # Communication Busy

    # parsing response
    message = message.split("!")[0]
    parameters = message[:-4].split("=")[1].split(",")

    error_code = int(str(parameters[3]))  # 0 is good
    if error_code != 0:
      raise RuntimeError(f"An error occurred: response {message}")

    sensor_status = int(str(parameters[5]))

    return A4SBackend.Status(
      current_temperature=int(str(parameters[0])) / 10,
      system_status=A4SBackend.Status.SystemStatus(int(str(parameters[1]))),
      heater_block_status=A4SBackend.Status.HeaterBlockStatus(int(str(parameters[2]))),
      error_code=error_code,
      warning_code=int(str(parameters[4])),
      sensor_status=A4SBackend.Status.SensorStatus(
        shuttle_middle_sensor=sensor_status & 0x0001 != 0,
        shuttle_open_sensor=sensor_status & 0x0002 != 0,
        shuttle_close_sensor=sensor_status & 0x0004 != 0,
        clean_door_sensor=sensor_status & 0x0008 != 0,
        seal_roll_sensor=sensor_status & 0x0010 != 0,
        heater_motor_up_sensor=sensor_status & 0x0020 != 0,
        heater_motor_down_sensor=sensor_status & 0x0040 != 0,
        # no_connect = sensor_status & 0x0080 != 0,
      ),
      remaining_time=int(str(parameters[6])),
    )

  async def _wait_for_status(self, statuses: Set["A4SBackend.Status.SystemStatus"]) -> Status:
    start = time.time()
    while True:
      status = await self.get_status()

      if status.system_status == A4SBackend.Status.SystemStatus.error:
        raise RuntimeError(f"An error occurred: {status.error_code}")

      if status.system_status in statuses:
        return status

      if time.time() - start > self.timeout:
        raise TimeoutError("Timeout while waiting for response")

      await asyncio.sleep(0.01)

  async def send_command(self, command: str):
    # command accepted: *Y01PL!
    # Command index: 01
    await self.io.write(command.encode())
    await asyncio.sleep(0.1)

  async def seal(self, temperature: int, duration: float):
    await self.set_temperature(temperature)
    await self.set_time(duration)
    await self.send_command("*00GS=zz!")  # Command to conduct seal action
    await self._wait_for_status({A4SBackend.Status.SystemStatus.single_cycle})
    return await self._wait_for_status(
      {A4SBackend.Status.SystemStatus.idle, A4SBackend.Status.SystemStatus.finish}
    )

  async def _wait_for_temperature(self, degrees: float, timeout: float, tolerance: float = 0.5):
    start = time.time()
    while True:
      current_temperature = await self.get_temperature()
      if abs(current_temperature - degrees) < tolerance:
        break
      if time.time() - start > timeout:
        raise TimeoutError("Timeout while waiting for temperature")
      await asyncio.sleep(0.1)

  async def _wait_for_shuttle_open_sensor(
    self, shuttle_open: bool, timeout: float = 30.0
  ) -> Status:
    start = time.time()
    while True:
      status = await self.get_status()
      if status.sensor_status.shuttle_open_sensor == shuttle_open:
        return status
      if time.time() - start > timeout:
        raise TimeoutError("Timeout while waiting for shuttle open sensor")

  async def set_temperature(self, temperature: float):
    if not (50 <= temperature <= 200):
      raise ValueError("Temperature out of range. Please enter a value between 50 and 200.")
    command = f"*00DH={round(temperature):04d}zz!"
    await self.send_command(command)
    await self._wait_for_temperature(temperature, timeout=300)

  async def set_time(self, seconds: float):
    deciseconds = seconds * 10
    if not (0 <= deciseconds <= 9999):
      raise ValueError("Time out of range. Please enter a value between 0 and 9999.")
    command = f"*00DT={deciseconds:04d}zz!"
    return await self.send_command(command)

  async def open(self) -> Status:
    await self.send_command("*00MO=zz!")
    return await self._wait_for_shuttle_open_sensor(True)

  async def close(self) -> Status:
    await self.send_command("*00MC=zz!")
    return await self._wait_for_shuttle_open_sensor(False)

  async def system_reset(self):
    await self.send_command("*00SR=zz!")
    return await self._wait_for_status({A4SBackend.Status.SystemStatus.idle})

  async def get_temperature(self) -> float:
    status = await self.get_status()
    return status.current_temperature

  async def get_remaining_time(self) -> int:
    status = await self.get_status()
    return status.remaining_time
